/*
 * Copyright (c) 2020 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

/*
 * ADIF 3.1.1
 *
 * http://www.adif.org/311/ADIF_311.htm
 */

#include <jeffpc/buffer.h>

#include "qso.h"
#include "qso_impl.h"
#include "bands.h"
#include "util.h"
#include "version.h"

#define ADIF_VERSION "3.1.1"

enum adif_type {
	AT_UUID,
	AT_DATE,
	AT_TIME,
	AT_STR,
	AT_INT,
	AT_POWER,
	AT_FREQ,
	AT_BAND,
};

struct adif_field {
	const char *name;
	enum adif_type type;
	enum qso_item what;
	ptrdiff_t qso_offset_if_diff;
	ptrdiff_t qso_offset;
};

static const struct adif_field fields[] = {
	{ "app_hlog_uuid", AT_UUID, QSO_UUID, -1,
		offsetof(struct qso, uuid), },
	{ "qso_date", AT_DATE, QSO_START, -1,
		offsetof(struct qso, start), },
	{ "qso_date_off", AT_DATE, QSO_END, -1,
		offsetof(struct qso, end), },
	{ "time_on", AT_TIME, QSO_START, -1,
		offsetof(struct qso, start), },
	{ "time_off", AT_TIME, QSO_END, -1,
		offsetof(struct qso, end), },
	{ "comment", AT_STR, QSO_COMMENT, -1,
		offsetof(struct qso, comment), },

	/* local / tx */
	{ "operator", AT_STR, QSO_TX_OPERATOR_CALL, -1,
		offsetof(struct qso, tx.operator_call), },
	{ "station_callsign", AT_STR, QSO_TX_STATION_CALL, -1,
		offsetof(struct qso, tx.station_call), },
	{ "tx_power", AT_POWER, QSO_TX_POWER, -1,
		offsetof(struct qso, tx.power), },
	{ "freq", AT_FREQ, QSO_TX_FREQ, -1,
		offsetof(struct qso, tx.freq), },
	{ "band", AT_BAND, QSO_NOT_A_THING, -1,
		offsetof(struct qso, tx.freq), },
	/* TODO: split mode into mode/submode */
	{ "mode", AT_STR, QSO_TX_MODE, -1,
		offsetof(struct qso, tx.mode), },
	{ "rst_sent", AT_STR, QSO_TX_RST, -1,
		offsetof(struct qso, tx.rst), },
	{ "my_name", AT_STR, QSO_TX_NAME, -1,
		offsetof(struct qso, tx.name), },
	{ "app_hlog_my_qth", AT_STR, QSO_TX_QTH, -1,
		offsetof(struct qso, tx.qth), },
	{ "my_gridsquare", AT_STR, QSO_TX_GRID, -1,
		offsetof(struct qso, tx.grid), },
	{ "my_country", AT_STR, QSO_TX_COUNTRY, -1,
		offsetof(struct qso, tx.country), },
	{ "my_state", AT_STR, QSO_TX_STATE, -1,
		offsetof(struct qso, tx.state), },
	{ "my_cnty", AT_STR, QSO_TX_COUNTY, -1,
		offsetof(struct qso, tx.county), },
	{ "app_hlog_my_cont", AT_STR, QSO_TX_CONTINENT, -1,
		offsetof(struct qso, tx.continent), },
	{ "app_hlog_my_dxcc", AT_INT, QSO_TX_DXCC, -1,
		offsetof(struct qso, tx.dxcc), },
	{ "my_itu_zone", AT_INT, QSO_TX_ITU, -1,
		offsetof(struct qso, tx.itu), },
	{ "my_cq_zone", AT_INT, QSO_TX_CQZ, -1,
		offsetof(struct qso, tx.cqz), },
	{ "stx", AT_INT, QSO_TX_SEQ, -1,
		offsetof(struct qso, tx.seq), },

	/* remote / rx */
	{ "call", AT_STR, QSO_RX_STATION_CALL, -1,
		offsetof(struct qso, rx.station_call), },
	{ "contacted_op", AT_STR, QSO_RX_OPERATOR_CALL,
		offsetof(struct qso, rx.station_call),
		offsetof(struct qso, rx.operator_call), },
	{ "rx_power", AT_POWER, QSO_RX_POWER, -1,
		offsetof(struct qso, rx.power), },
	{ "freq_rx", AT_FREQ, QSO_RX_FREQ,
		offsetof(struct qso, tx.freq),
		offsetof(struct qso, rx.freq), },
	{ "band_rx", AT_BAND, QSO_NOT_A_THING,
		offsetof(struct qso, tx.freq),
		offsetof(struct qso, rx.freq), },
	/* TODO: split mode into mode/submode */
	{ "app_hlog_mode_rx", AT_STR, QSO_RX_MODE,
		offsetof(struct qso, tx.mode),
		offsetof(struct qso, rx.mode), },
	{ "rst_rcvd", AT_STR, QSO_RX_RST, -1,
		offsetof(struct qso, rx.rst), },
	{ "name", AT_STR, QSO_RX_NAME, -1,
		offsetof(struct qso, rx.name), },
	{ "qth", AT_STR, QSO_RX_QTH, -1,
		offsetof(struct qso, rx.qth), },
	{ "gridsquare", AT_STR, QSO_RX_GRID, -1,
		offsetof(struct qso, rx.grid), },
	{ "country", AT_STR, QSO_RX_COUNTRY, -1,
		offsetof(struct qso, rx.country), },
	{ "state", AT_STR, QSO_RX_STATE, -1,
		offsetof(struct qso, rx.state), },
	{ "cnty", AT_STR, QSO_RX_COUNTY, -1,
		offsetof(struct qso, rx.county), },
	{ "cont", AT_STR, QSO_RX_CONTINENT, -1,
		offsetof(struct qso, rx.continent), },
	{ "dxcc", AT_INT, QSO_RX_DXCC, -1,
		offsetof(struct qso, rx.dxcc), },
	{ "ituz", AT_INT, QSO_RX_ITU, -1,
		offsetof(struct qso, rx.itu), },
	{ "cqz", AT_INT, QSO_RX_CQZ, -1,
		offsetof(struct qso, rx.cqz), },
	{ "srx", AT_INT, QSO_RX_SEQ, -1,
		offsetof(struct qso, rx.seq), },
};

static int __adif_write_cstr(struct buffer *buf, const char *name,
			     const char *value)
{
	const size_t len = value ? strlen(value) : 0;
	char lenbuf[25];
	int ret;

	if (!len)
		return 0;

	snprintf(lenbuf, sizeof(lenbuf), ":%zu>", len);

	if ((ret = buffer_append_c(buf, '<')))
		return ret;
	if ((ret = buffer_append_cstr(buf, name)))
		return ret;
	if ((ret = buffer_append_cstr(buf, lenbuf)))
		return ret;
	if (len && ((ret = buffer_append(buf, value, len))))
		return ret;
	if ((ret = buffer_append_c(buf, '\n')))
		return ret;

	return 0;
}

static int __adif_write_str(struct buffer *buf, const char *name,
			    const struct str *value)
{
	return __adif_write_cstr(buf, name, str_cstr(value));
}

static int __adif_write_uuid(struct buffer *buf, const char *name,
			     const struct xuuid *uuid)
{
	char str[XUUID_PRINTABLE_STRING_LENGTH];

	ASSERT(!xuuid_is_null(uuid));

	xuuid_unparse(uuid, str);

	return __adif_write_cstr(buf, name, str);
}

static int __adif_write_int(struct buffer *buf, const char *name,
			    struct val *value)
{
	char str[30];

	if (!value)
		return 0;

	snprintf(str, sizeof(str), "%"PRIu64, value->i);

	return __adif_write_cstr(buf, name, str);
}

static int __adif_write_float(struct buffer *buf, const char *name,
			      struct val *value, int scale, int digits)
{
	char str[30];
	uint64_t whole;
	uint64_t frac;

	if (!value)
		return 0;

	whole = value->i / scale;
	frac = value->i % scale;

	while (frac && ((frac % 10) == 0)) {
		frac /= 10;
		digits--;
	}

	if (frac)
		snprintf(str, sizeof(str), "%"PRIu64".%0*"PRIu64, whole, digits,
			 frac);
	else
		snprintf(str, sizeof(str), "%"PRIu64, whole);

	return __adif_write_cstr(buf, name, str);
}

static int __adif_write_date(struct buffer *buf, const char *name,
			     const struct time *time)
{
	char str[9];

	if (!(time->have & TIME_PART_DATE))
		return 0;

	snprintf(str, sizeof(str), "%04u%02u%02u",
		 time->year, time->month, time->day);

	return __adif_write_cstr(buf, name, str);
}

static int __adif_write_time(struct buffer *buf, const char *name,
			     const struct time *time)
{
	char str[7];

	if (!(time->have & TIME_PART_TIME_HM))
		return 0;

	if (time->have & TIME_PART_TIME_S)
		snprintf(str, sizeof(str), "%02u%02u%02u",
			 time->hour, time->min, time->sec);
	else
		snprintf(str, sizeof(str), "%02u%02u",
			 time->hour, time->min);

	return __adif_write_cstr(buf, name, str);
}

static int adif_write_additional(struct buffer *buf, const struct qso *qso)
{
	const struct nvpair *pair;
	const char *name;
	struct str *value;
	char tmp[20];
	int ret;

	if (!qso->additional)
		return 0;

	nvl_for_each(pair, qso->additional) {
		name = nvpair_name(pair);
		value = val_cast_to_str(nvpair_value(pair));

		/*
		 * TODO: The above cast may panic if the value is not a
		 * string.  This cannot happen with ADIF-imported QSOs, but
		 * may happen if the additional nvlist contains anything
		 * other than strings.
		 */

		snprintf(tmp, sizeof(tmp), ":%zu>", str_len(value));

		/* append the key: <name:len> */
		ret = buffer_append_c(buf, '<');
		if (ret)
			goto err_putref;

		ret = buffer_append_cstr(buf, name);
		if (ret)
			goto err_putref;

		ret = buffer_append_cstr(buf, tmp);
		if (ret)
			goto err_putref;

		/* append the value: blob */
		ret = buffer_append_str(buf, value);
		if (ret)
			goto err_putref;

		/* append a newline for clarity */
		ret = buffer_append_c(buf, '\n');
		if (ret)
			goto err_putref;

		str_putref(value);
	}

	return 0;

err_putref:
	str_putref(value);

	return ret;
}

int adif_write(struct buffer *buf, const struct qso *qso)
{
	size_t i;
	int ret;

	/* header */
	if ((ret = buffer_append_cstr(buf, "###\n")))
		return ret;
	if ((ret = __adif_write_cstr(buf, "adif_ver", ADIF_VERSION)))
		return ret;
	if ((ret = __adif_write_cstr(buf, "programid", "hlog")))
		return ret;
	if ((ret = __adif_write_cstr(buf, "programversion", version_string)))
		return ret;
	if ((ret = buffer_append_cstr(buf, "<eoh>\n")))
		return ret;

	/* body */
	for (i = 0; i < ARRAY_LEN(fields); i++) {
		const struct adif_field *field = &fields[i];
		const uintptr_t base = (uintptr_t) qso;
		const void *ptr = (void *)(base + field->qso_offset);
		const void *ifdiff = (void *)(base + field->qso_offset_if_diff);
		const bool diff_only = field->qso_offset_if_diff != -1;

		ret = -EINVAL;

		switch (field->type) {
			case AT_UUID:
				ASSERT(!diff_only);
				ret = __adif_write_uuid(buf, field->name, ptr);
				break;
			case AT_DATE:
				ASSERT(!diff_only);
				ret = __adif_write_date(buf, field->name, ptr);
				break;
			case AT_TIME:
				ASSERT(!diff_only);
				ret = __adif_write_time(buf, field->name, ptr);
				break;
			case AT_STR: {
				struct str **this = (struct str **) ptr;
				struct str **other = (struct str **) ifdiff;

				if (diff_only &&
				    ((!*this && !*other) ||
				     (*this && *other &&
				      !strcmp(str_cstr(*this), str_cstr(*other))))) {
					ret = 0;
					break;
				}

				ret = __adif_write_str(buf, field->name, *this);
				break;
			}
			case AT_INT:
				ASSERT(!diff_only);
				ret = __adif_write_int(buf, field->name,
						       *(struct val **) ptr);
				break;
			case AT_POWER:
				ASSERT(!diff_only);
				ret = __adif_write_float(buf, field->name,
							 *(struct val **) ptr,
							 1000, 3);
				break;
			case AT_FREQ: {
				struct val **this = (struct val **) ptr;
				struct val **other = (struct val **) ifdiff;

				if (diff_only &&
				    ((!*this && !*other) ||
				     (*this && *other &&
				      ((*this)->i == (*other)->i)))) {
					ret = 0;
					break;
				}

				ret = __adif_write_float(buf, field->name,
							 *this, 1000000, 6);
				break;
			}
			case AT_BAND: {
				struct val **this = (struct val **) ptr;
				struct val **other = (struct val **) ifdiff;
				struct str *band;

				ret = 0;

				if (diff_only &&
				    ((!*this && !*other) ||
				     (*this && *other &&
				      ((*this)->i == (*other)->i))))
					break;

				if (!*this)
					break;

				band = bands_lookup_adif_name_by_freq((*this)->i);
				if (!band)
					break;

				ret = __adif_write_str(buf, field->name, band);

				str_putref(band);
				break;
			}
		}

		if (ret)
			return ret;
	}

	/*
	 * Write out any additional fields that aren't part of the struct
	 * qso itself.
	 */
	ret = adif_write_additional(buf, qso);
	if (ret)
		return ret;

	return buffer_append_cstr(buf, "<eor>\n");
}

static inline int get_byte(struct buffer *buf, uint8_t *byte)
{
	ssize_t ret;

	ret = buffer_read(buf, byte, sizeof(*byte));
	if (ret < 0)
		return ret;
	if (!ret)
		return -EEOF;
	return 0;
}

static inline int skip_until_le(struct buffer *buf)
{
	uint8_t byte;
	int ret;

	do {
		ret = get_byte(buf, &byte);
		if (ret)
			return ret;
	} while (byte != '<');

	return 0;
}

static int consume_header(struct buffer *input)
{
	uint8_t byte;
	char last[4];
	ssize_t ret;

	ret = get_byte(input, &byte);
	if (ret)
		return ret;

	if (byte == '<')
		return 0; /* no header */

	for (;;) {
		/* find the next field */
		ret = skip_until_le(input);
		if (ret)
			return ret;

		/* read the name */
		ret = buffer_read(input, last, sizeof(last));
		if (ret < 0)
			return ret;
		if (ret != sizeof(last))
			return -EEOF;

		/* if it is <eoh>, we're almost done */
		if (!strncmp(last, "eoh>", sizeof(last)))
			break;
	}

	/*
	 * Skip under the next '<' to leave the input in the same state
	 * regardless of if there was a header or not - with the first
	 * non-header field '<' already consumed.
	 */
	return skip_until_le(input);
}

static int interpret_field(struct qso *qso,
			   struct buffer *namebuf,
			   struct buffer *valuebuf)
{
	size_t value_len;
	size_t name_len;
	char *value;
	char *name;
	size_t i;

	name = buffer_data_mutable(namebuf);
	name_len = buffer_size(namebuf) - 1;
	if (strlen(name) != name_len) {
		fprintf(stderr, "Warning: not interpreting field "
			"name with nul ('%s'; %zu != %zu)\n",
			name, strlen(name), name_len);
		return 0;
	}

	value = buffer_data_mutable(valuebuf);
	value_len = buffer_size(valuebuf) - 1;
	if (strlen(value) != value_len) {
		fprintf(stderr, "Warning: not interpreting field "
			"value with nul ('%s'; %zu != %zu)\n",
			value, strlen(value), value_len);
		return 0;
	}

	for (i = 0; i < ARRAY_LEN(fields); i++) {
		const struct adif_field *field = &fields[i];

		if (strcasecmp(name, field->name))
			continue;

		/* found */
		return qso_set(qso, field->what, value);
	}

	/* stash unknown fields in struct qso's additional nvlist */

	lowercase_inplace(name);

	return nvl_set_cstr_dup(qso->additional, name, value);
}

int adif_parse(struct buffer *buf, struct qso *qso)
{
	struct buffer *value;
	struct buffer name;
	char rawname[64];
	uint8_t byte;
	ssize_t ret;
	size_t len;

	qso_reset(qso, true);

	buffer_init_static(&name, rawname, sizeof(rawname), sizeof(rawname),
			   true);

	value = buffer_alloc(1024);
	if (IS_ERR(value))
		return PTR_ERR(value);

	ret = consume_header(buf);
	if (ret)
		goto err;

	for (;;) {
		/* empty name & value buffers */
		VERIFY0(buffer_truncate(&name, 0));
		VERIFY0(buffer_truncate(value, 0));

		/* read field name until we get a ':' or '>' */
		for (;;) {
			ret = get_byte(buf, &byte);
			if (ret)
				goto err;

			if ((byte == ':') || (byte == '>'))
				break;

			ret = buffer_append_c(&name, byte);
			if (ret)
				goto err;
		}

		/* nul-terminate the name */
		ret = buffer_append_c(&name, '\0');
		if (ret)
			goto err;

		/* read value length until we get a ':' or '>' */
		len = 0;
		if (byte != '>') {
			for (;;) {
				ret = get_byte(buf, &byte);
				if (ret)
					goto err;

				if ((byte == ':') || (byte == '>'))
					break;

				if ((byte < '0') || (byte > '9')) {
					ret = -EINVAL;
					goto err;
				}

				len = (len * 10) + (byte - '0');
			}

			/* skip until we get a '>' */
			if (byte == ':') {
				do {
					ret = get_byte(buf, &byte);
					if (ret)
						goto err;
				} while (byte != '>');
			}
		}

		if (!strcasecmp(buffer_data(&name), "eor")) {
			ret = 0; /* yes, a success */
			goto err;
		}

		/* read in the value */
		ret = buffer_copy(value, buf, len);
		if (ret < 0)
			goto err;

		/* nul-terminate the value */
		ret = buffer_append_c(value, '\0');
		if (ret)
			goto err;

		/* interpret the field name-value */
		ret = interpret_field(qso, &name, value);
		if (ret)
			goto err;

		/* find the next field */
		ret = skip_until_le(buf);
		if (ret)
			return ret;
	}

	/* TODO: validate the qso */
	/* TODO:  - check that we actually parsed a uuid */
	ASSERT(!xuuid_is_null(&qso->uuid));

	ret = 0;

err:
	buffer_free(value);

	return ret;
}
