/*
 * Copyright (c) 2020 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <jeffpc/list.h>
#include <jeffpc/sexpr.h>
#include <jeffpc/io.h>
#include <jeffpc/mem.h>

#include "bands.h"

struct band {
	struct list_node list;

	struct str *name_iaru;
	struct str *name_adif;
	uint64_t name_iaru_mm;
	uint64_t f_start;
	uint64_t f_end;
};

static struct list bands;

static struct band *lookup_band_by_freq(uint64_t freq)
{
	struct band *band;

	list_for_each(band, &bands) {
		if ((freq < band->f_start) || (freq > band->f_end))
			continue;

		return band;
	}

	return NULL;
}

struct str *bands_lookup_name_by_freq(uint64_t freq)
{
	struct band *band;

	band = lookup_band_by_freq(freq);

	return band ? str_getref(band->name_iaru) : NULL;
}

struct str *bands_lookup_adif_name_by_freq(uint64_t freq)
{
	struct band *band;

	band = lookup_band_by_freq(freq);

	return band ? str_getref(band->name_adif) : NULL;
}

uint64_t bands_lookup_lambda_by_freq(uint64_t freq)
{
	struct band *band;

	band = lookup_band_by_freq(freq);

	return band ? band->name_iaru_mm : 0;
}

static struct band *cvt_band(struct val *parsed)
{
#define NUM_ITEMS 5
	struct val *items[NUM_ITEMS];
	struct band *band;
	size_t i;

	if (sexpr_list_to_array(parsed, items, NUM_ITEMS) != NUM_ITEMS)
		return ERR_PTR(-EINVAL);

	if ((items[0]->type != VT_STR) || (items[1]->type != VT_INT) ||
	    (items[2]->type != VT_STR) || (items[3]->type != VT_INT) ||
	    (items[4]->type != VT_INT)) {
		band = ERR_PTR(-EINVAL);
		goto err;
	}

	band = zalloc(sizeof(struct band));
	if (!band) {
		band = ERR_PTR(-ENOMEM);
		goto err;
	}

	band->name_iaru = val_getref_str(items[0]);
	band->name_iaru_mm = items[1]->i;
	band->name_adif = val_getref_str(items[2]);
	band->f_start = items[3]->i;
	band->f_end = items[4]->i;

err:
	for (i = 0; i < ARRAY_LEN(items); i++)
		val_putref(items[i]);

	return band;
}

int bands_load(int dirfd, const char *fname)
{
	struct val *cur, *tmp;
	struct val *parsed;
	const char *raw;
	size_t len;
	int ret;

	list_create(&bands, sizeof(struct band), offsetof(struct band, list));

	raw = read_file_len_at(dirfd, fname, &len);
	if (IS_ERR(raw))
		return PTR_ERR(raw);

	parsed = sexpr_parse(raw, len);

	/* free it now so we don't have to worry about it anymore */
	free((void *) raw);

	if (IS_ERR(parsed))
		return PTR_ERR(parsed);

	sexpr_for_each_noref(cur, tmp, parsed) {
		struct band *band;

		band = cvt_band(cur);
		if (IS_ERR(band)) {
			ret = PTR_ERR(band);
			goto err;
		}

		printf("band: %-5s %11.6f %11.6f\n",
		       str_cstr(band->name_iaru),
		       band->f_start / 1e6,
		       band->f_end / 1e6);

		/* TODO: replace with an interval tree to check for overlap */
		list_insert_tail(&bands, band);
	}

	val_putref(parsed);

	return 0;

err:
	val_putref(parsed);

	/* TODO: free already allocated bands */

	return ret;
}
