/*
 * Copyright (c) 2020 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <jeffpc/version.h>
#include <jeffpc/error.h>
#include <jeffpc/time.h>
#include <jeffpc/val.h>
#include <jeffpc/sexpr.h>
#include <jeffpc/io.h>
#include <jeffpc/unicode.h>
#include <jeffpc/cstr.h>

#include <ctype.h>
#include <locale.h>
#include <ncurses.h>
#include <math.h>
#include <spawn.h>

#include "version.h"
#include "qso.h"
#include "adif.h"
#include "bands.h"
#include "index.h"
#include "country.h"
#include "maidenhead.h"
#include "util.h"

#define EARTH_RADIUS		6378.137

#define ENTRY_WIN_HEIGHT	(6 + 1)
#define ENTRY_WIN_WIDTH		80
#define ENTRY_WIN_ROW		0
#define CALL_WIN_HEIGHT		10
#define CALL_WIN_WIDTH		ENTRY_WIN_WIDTH
#define CALL_WIN_ROW		(ENTRY_WIN_ROW + ENTRY_WIN_HEIGHT)
#define CMD_WIN_HEIGHT		3
#define CMD_WIN_WIDTH		RECENT_WIN_WIDTH
#define CMD_WIN_ROW		(LINES - 3)
#define RECENT_WIN_HEIGHT	(LINES - ENTRY_WIN_HEIGHT - CALL_WIN_HEIGHT - \
				 CMD_WIN_HEIGHT)
#define RECENT_WIN_WIDTH	CALL_WIN_WIDTH
#define RECENT_WIN_ROW		(CALL_WIN_ROW + CALL_WIN_HEIGHT)

#define CMD_HINT_COLOR		1

#define MAX_STATE_COUNT		10 /* number of concurrent states (tabs) */
#define MAX_SUMMARY_MODE_COUNT	4 /* number of modes to display in call summary */
#define MAX_SUMMARY_BAND_COUNT	14 /* number of bands to display in call summary */

#define DATA_DIR_CONFIG_FNAME	"config.scm"
#define DATA_DIR_COUNTRY_FNAME	"cty.csv"
#define DATA_DIR_BAND_FNAME	"bands.scm"

static const char *data_dir;
static int data_dir_fd;

static struct str default_date = STR_STATIC_INITIALIZER("????""-??""-??");
static struct str default_time = STR_STATIC_INITIALIZER("??:??:??");
static struct str default_latlon = STR_STATIC_INITIALIZER("??.??N ???.??E");
static struct str default_azidist =
	STR_STATIC_INITIALIZER("???° ????? km/???° ????? km");

/* various parts of the display */
enum refresh {
	REF_DE =		0x00000001,
	REF_DE_GRID =		0x00000002,
	REF_DE_POWER =		0x00000004,
	REF_DATE =		0x00000008,
	REF_TIME_START =	0x00000010,
	REF_TIME_END =		0x00000020,
	REF_CALL =		0x00000040,
	REF_FREQ =		0x00000080,
	REF_MODE =		0x00000100,
	REF_RSTS =		0x00000200,
	REF_RSTR =		0x00000400,
	REF_NAME =		0x00000800,
	REF_QTH =		0x00001000,
	REF_GRID =		0x00002000,
	REF_COUNTRY =		0x00004000,
	REF_GEO =		0x00008000,
	REF_STATE =		0x00010000,
	REF_COUNTY =		0x00020000,
	REF_ZONE_CONT =		0x00040000,
	REF_ZONE_DXCC =		0x00080000,
	REF_ZONE_ITU =		0x00100000,
	REF_ZONE_CQZ =		0x00200000,
#define REF_ZONES	(REF_ZONE_CONT | REF_ZONE_DXCC | REF_ZONE_ITU | \
			 REF_ZONE_CQZ)
	REF_COMMENT =		0x00400000,
	REF_SEQ =		0x00800000,
	REF_CALL_WIN =		0x40000000,
	REF_RECENT_WIN =	0x80000000,
};

/* This contains everything being displayed on screen */
struct state {
	struct qso qso;

	/* the following are generated from the QSO data above */
	struct str *date;
	struct str *time_start;
	struct str *time_end;
	struct str *latlon; /* lat & lon */
	struct str *azidist; /* azimuth & distance */

	enum refresh dirty; /* which fields have been modified since last save */
	bool in_use; /* is this state in-use */
};

struct stats_summary {
	struct {
		struct str *mode_name;
		uint64_t band[MAX_SUMMARY_BAND_COUNT];
		uint64_t total; /* number of time this mode has been used */
	} mode[MAX_SUMMARY_MODE_COUNT];
};

/* list of bands (in mm) in call summary table */
static uint64_t summary_bands[MAX_SUMMARY_BAND_COUNT] = {
	0,
	160000,
	80000,
	60000,
	40000,
	30000,
	20000,
	17000,
	15000,
	12000,
	10000,
	6000,
	2000,
	700,
};

static struct state *current;
static struct state states[MAX_STATE_COUNT];

struct entry_field {
	int row;
	int col_label;
	int col_value;
	/*
	 * The field label.  All characters other than '_' are printed.  A
	 * '_' indicates that the following character should be highlighted
	 * to hint at the command to modify this field.
	 */
	const char *label;
	size_t len;
	enum val_type type;
	enum refresh refresh;
	ptrdiff_t offset;
};

enum cmd_result {
	CR_OK = 0,		/* success */
	CR_INVAL_CMD,		/* unknown command */
	CR_MISSING_ARG,		/* command requires an arg */
	CR_EXTRA_ARG,		/* command requires no args */
	CR_TOO_MANY_TABS,	/* cannot open new tab */
	CR_INVAL_TAB_ID,	/* malformed tab id */
	CR_UNKNOWN_TAB_ID,	/* unknown tab id */
};

struct command {
	const char *cmd;
	enum cmd_result (*fxn)(char *arg, char *error, size_t errorlen);
	bool arg_required;
	enum refresh refresh;
};

static enum cmd_result cmd_tab(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_new(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_reset(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_save(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_edit(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_start(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_end(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_qrz(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_de(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_de_grid(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_de_power(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_call(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_freq(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_mode(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_rst_s(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_rst_r(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_name(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_qth(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_grid(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_country(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_state(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_county(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_continent(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_dxcc(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_itu(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_cqz(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_comment(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_seq_r(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_seq_s(char *arg, char *error, size_t errorlen);

static const struct entry_field entry_fields[] = {
	{ 0,  0,  3, "_d_e",          10, VT_STR, REF_DE,
	  offsetof(struct state, qso.tx.station_call), },
	{ 0, 15, 21, "_Grid:",         8, VT_STR, REF_DE_GRID,
	  offsetof(struct state, qso.tx.grid), },
	{ 0, 31, 36, "_P_w_r:",        5, VT_INT, REF_DE_POWER,
	  offsetof(struct state, qso.tx.power), },
	{ 0,  0, 52, NULL,            10, VT_STR, REF_DATE,
	  offsetof(struct state, date), },
	{ 0,  0, 63, NULL,             8, VT_STR, REF_TIME_START,
	  offsetof(struct state, time_start), },
	{ 0,  0, 72, NULL,             8, VT_STR, REF_TIME_END,
	  offsetof(struct state, time_end), },
	{ 1,  0,  6, "_Call:",        10, VT_STR, REF_CALL,
	  offsetof(struct state, qso.rx.station_call), },
	{ 1, 18, 24, "_Freq:",        10, VT_INT, REF_FREQ,
	  offsetof(struct state, qso.tx.freq), },
	{ 1, 36, 42, "_Mode:",         6, VT_STR, REF_MODE,
	  offsetof(struct state, qso.tx.mode), },
	{ 1, 50, 56, "_RST_s:",        3, VT_STR, REF_RSTS,
	  offsetof(struct state, qso.tx.rst), },
	{ 1, 61, 67, "_RST_r:",        3, VT_STR, REF_RSTR,
	  offsetof(struct state, qso.rx.rst), },
	{ 2,  0,  6, "_Na_me:",       16, VT_STR, REF_NAME,
	  offsetof(struct state, qso.rx.name), },
	{ 2, 24, 29, "_Q_T_H:",       35, VT_STR, REF_QTH,
	  offsetof(struct state, qso.rx.qth), },
	{ 2, 66, 72, "_Grid:",         8, VT_STR, REF_GRID,
	  offsetof(struct state, qso.rx.grid), },
	{ 3,  0,  9, "_C_o_u_n_t_r_y:", 26, VT_STR, REF_COUNTRY,
	  offsetof(struct state, qso.rx.country), },
	{ 3,  0, 37, NULL,            14, VT_STR, REF_GEO,
	  offsetof(struct state, latlon), },
	{ 3,  0, 53, NULL,            27, VT_STR, REF_GEO,
	  offsetof(struct state, azidist), },
	{ 4,  0,  7, "_S_t/Pr:",       2, VT_STR, REF_STATE,
	  offsetof(struct state, qso.rx.state), },
	{ 4, 11, 19, "_C_o_u_n_t_y:", 22, VT_STR, REF_COUNTY,
	  offsetof(struct state, qso.rx.county), },
	{ 4, 43, 49, "_C_o_n_t:",      2, VT_STR, REF_ZONE_CONT,
	  offsetof(struct state, qso.rx.continent), },
	{ 4, 53, 59, "_D_X_C_C:",      3, VT_INT, REF_ZONE_DXCC,
	  offsetof(struct state, qso.rx.dxcc), },
	{ 4, 64, 69, "_I_T_U:",        2, VT_INT, REF_ZONE_ITU,
	  offsetof(struct state, qso.rx.itu), },
	{ 4, 73, 78, "_C_Q_Z:",        2, VT_INT, REF_ZONE_CQZ,
	  offsetof(struct state, qso.rx.cqz), },
	{ 5,  0,  9, "_Co_mment:",    71, VT_STR, REF_COMMENT,
	  offsetof(struct state, qso.comment), },
};

static const struct command commands[] = {
	{ "tab",    cmd_tab,		true,  ~0u },
	{ "new",    cmd_new,		false, ~0u },
	{ "reset",  cmd_reset,		false, ~0u },
	{ "save",   cmd_save,		false, ~0u /* clears dirty bits */ },
	{ "edit",   cmd_edit,		true,  ~0u },
	{ "start",  cmd_start,		true,  REF_DATE | REF_TIME_START, },
	{ "end",    cmd_end,		true,  REF_DATE | REF_TIME_END, },
	{ "qrz",    cmd_qrz,		false, 0 },

	{ "de",     cmd_de,		true,  REF_DE, },
	{ "dg",     cmd_de_grid,	true,  REF_DE_GRID | REF_GEO, },
	{ "dpwr",   cmd_de_power,	true,  REF_DE_POWER, },
	{ "c",      cmd_call,		true,  REF_CALL | REF_COUNTRY |
					       REF_GEO | REF_ZONES |
					       REF_CALL_WIN, },
	{ "f",      cmd_freq,		true,  REF_FREQ, },
	{ "m",      cmd_mode,		true,  REF_MODE, },
	{ "rs",     cmd_rst_s,		true,  REF_RSTS, },
	{ "rr",     cmd_rst_r,		true,  REF_RSTR, },
	{ "nm",     cmd_name,		true,  REF_NAME, },
	{ "qth",    cmd_qth,		true,  REF_QTH, },
	{ "g",      cmd_grid,		true,  REF_GRID | REF_GEO, },
	{ "country",cmd_country,	true,  REF_COUNTRY, },
	{ "st",     cmd_state,		true,  REF_STATE, },
	{ "county", cmd_county,		true,  REF_COUNTY, },
	{ "cont",   cmd_continent,	true,  REF_ZONE_CONT, },
	{ "dxcc",   cmd_dxcc,		true,  REF_ZONE_DXCC, },
	{ "itu",    cmd_itu,		true,  REF_ZONE_ITU, },
	{ "cqz",    cmd_cqz,		true,  REF_ZONE_CQZ, },
	{ "cm",     cmd_comment,	true,  REF_COMMENT, },
	{ "seqr",   cmd_seq_r,		true,  0 },
	{ "seqs",   cmd_seq_s,		true,  0 },
};

static WINDOW *entry_win;
static WINDOW *call_win;
static WINDOW *recent_win;
static WINDOW *cmd_win;

/* draw a horizontal line at specified row */
static void whrule(WINDOW *win, int row, int width)
{
	mvwhline(win, row, 0, 0, width);
}

/* print width amount of chars of string, padding or truncating as necessary */
static inline void wprintw_cstr(WINDOW *win, const char *str, size_t width)
{
	size_t inlen;

	if (!str)
		str = "";

	inlen = strlen(str);

	if (inlen >= width)
		wprintw(win, "%-*.*s ", width, width, str);
	else
		wprintw(win, "%-*s ", width, str);
}

/*
 * Command handler helpers
 */
/* returns NULL on error */
static struct state *tab_alloc(void)
{
	int tab;

	for (tab = 0; tab < ARRAY_LEN(states); tab++) {
		if (states[tab].in_use)
			continue;

		states[tab].in_use = true;
		return &states[tab];
	}

	return NULL;
}

static void tab_close(struct state *tab)
{
	tab->in_use = false;
}

/* returns NULL on error */
static struct state *tab_find_first_inuse(void)
{
	int tab;

	for (tab = 0; tab < ARRAY_LEN(states); tab++) {
		if (!states[tab].in_use)
			continue;

		return &states[tab];
	}

	return NULL;
}

/* returns negated errno on error */
static struct state *tab_find_id(const char *str)
{
	uint32_t tab;
	int ret;

	ret = str2u32(str, &tab);
	if (ret)
		return ERR_PTR(ret);

	if (tab >= MAX_STATE_COUNT)
		return ERR_PTR(-EINVAL);

	return &states[tab];
}

static void __activate_tab(struct state *tab)
{
	ASSERT(tab->in_use);
	current = tab;
}

static void __recalculate_time(const struct time *time, struct str **out)
{
	char buf[9];

	if (time_snprint_time(buf, sizeof(buf), time))
		*out = STR_DUP(buf);
	else
		*out = &default_time;
}

static void recalculate_times(void)
{
	struct time *date;
	char buf[11];

	str_putref(current->date);
	str_putref(current->time_start);
	str_putref(current->time_end);

	if (current->qso.start.have & TIME_PART_DATE)
		date = &current->qso.start;
	else if (current->qso.end.have & TIME_PART_DATE)
		date = &current->qso.end;
	else
		date = NULL;

	if (time_snprint_date(buf, sizeof(buf), date))
		current->date = STR_DUP(buf);
	else
		current->date = &default_date;

	__recalculate_time(&current->qso.start, &current->time_start);
	__recalculate_time(&current->qso.end, &current->time_end);
}

static void recalculate_location(void)
{
	double lat, lon;

	str_putref(current->latlon);

	if (maidenhead_decode(str_cstr(current->qso.rx.grid), &lat, &lon) ||
	    country_latlon(current->qso.rx.station_call, &lat, &lon)) {
		char buf[15];
		char ns, ew;

		ns = (lat >= 0) ? 'N' : 'S';
		ew = (lon >= 0) ? 'E' : 'W';

		snprintf(buf, sizeof(buf), "%5.2f%c %6.2f%c",
			 fabs(lat), ns,
			 fabs(lon), ew);

		current->latlon = STR_DUP(buf);
	} else {
		current->latlon = &default_latlon;
	}
}

static void recalculate_distance(void)
{
	double de_lat, de_lon;
	double dx_lat, dx_lon;
	double distance;
	double azimuth;
	unsigned int int_azi;
	char buf[30];

	str_putref(current->azidist);

	if (!current->qso.tx.grid)
		goto not_calculated; /* no idea where we are */
	if (!current->qso.rx.station_call && !current->qso.rx.grid)
		goto not_calculated; /* no idea where they are */

	if (!maidenhead_decode(str_cstr(current->qso.tx.grid), &de_lat, &de_lon))
		goto not_calculated; /* no idea where we are */

	if (!maidenhead_decode(str_cstr(current->qso.rx.grid), &dx_lat, &dx_lon) &&
	    !country_latlon(current->qso.rx.station_call, &dx_lat, &dx_lon))
		goto not_calculated; /* no idea where they are */

	/* degrees to radians */
	de_lat *= M_PI / 180.;
	de_lon *= M_PI / 180.;
	dx_lat *= M_PI / 180.;
	dx_lon *= M_PI / 180.;

	/*
	 * The following assumes a spherical earth.  This makes all the
	 * numbers slightly off from what we'd get if we used WGS84
	 * math...which we should.
	 */

	/* calculate */
	distance = acos(sin(de_lat) * sin(dx_lat) +
			cos(de_lat) * cos(dx_lat) * cos(de_lon - dx_lon));
	azimuth = fmod(atan2(sin(dx_lon - de_lon) * cos(dx_lat),
			     cos(de_lat) * sin(dx_lat) -
			     sin(de_lat) * cos(dx_lat) * cos(dx_lon - de_lon)),
		       2 * M_PI);

	/* radians to km & degrees */
	distance *= (2 * M_PI * EARTH_RADIUS) / (2 * M_PI);
	azimuth *= 180 / M_PI;

	/* convert to an int */
	int_azi = (azimuth >= 0.) ? azimuth : (azimuth + 360.);

	snprintf(buf, sizeof(buf), "%3u° %5.0f km/%3u° %5.0f km",
		 int_azi, distance,
		 (int_azi < 180) ? (int_azi + 180) : (int_azi - 180),
		 2 * M_PI * EARTH_RADIUS - distance);

	current->azidist = STR_DUP(buf);
	return;

not_calculated:
	current->azidist = &default_azidist;
}

/*
 * Command handlers
 */
static enum cmd_result cmd_tab(char *arg, char *error, size_t errorlen)
{
	struct state *new;

	if (!strcasecmp(arg, "new")) {
		/* create a new tab */
		new = tab_alloc();
		if (!new)
			return CR_TOO_MANY_TABS;

		__activate_tab(new);
	} else if (!strcasecmp(arg, "close")) {
		/* close an existing tab */
		tab_close(current);

		new = tab_find_first_inuse();
		if (new) {
			__activate_tab(new);
			return CR_OK;
		}

		/* all tabs closed, open a new one */

		new = tab_alloc(); /* cannot fail, all tabs closed */
		ASSERT3P(new, !=, NULL);

		__activate_tab(new);
	} else {
		/* switch to an existing tab */
		new = tab_find_id(arg);
		if (IS_ERR(new))
			return CR_INVAL_TAB_ID;
		if (!new->in_use)
			return CR_UNKNOWN_TAB_ID;

		__activate_tab(new);
		return CR_OK;
	}

	/* reset all the on-screen state to start from scratch */
	qso_reset(&current->qso, false);
	recalculate_times();
	recalculate_location();
	recalculate_distance();

	/* nothing is dirty - we reset it all */
	current->dirty = 0;

	return CR_OK;
}

static enum cmd_result cmd_new(char *arg, char *error, size_t errorlen)
{
	int ret;

	ret = qso_save(&current->qso, data_dir_fd);
	if (ret) {
		wprintw(recent_win, "Failed to save QSO: %s",
			xstrerror(ret));
		return ret; /* returning a negated errno */
	}

	ret = index_save(&current->qso);
	if (ret)
		wprintw(recent_win, "Warning: Failed to add QSO to index, "
			"UI may be out-of-sync: %s", xstrerror(ret));
	/* ignoring the index failure */

	/* display the uuid in the info message */
	snprintf(error, errorlen, "Saved %s", current->qso.uuid_str);

	qso_reset(&current->qso, false);

	recalculate_times();
	recalculate_location();
	recalculate_distance();

	current->dirty = 0;

	return CR_OK;
}

static enum cmd_result cmd_reset(char *arg, char *error, size_t errorlen)
{
	qso_reset(&current->qso, false);

	recalculate_times();
	recalculate_location();
	recalculate_distance();

	current->dirty = 0;

	return CR_OK;
}

static enum cmd_result cmd_save(char *arg, char *error, size_t errorlen)
{
	int ret;

	ret = qso_save(&current->qso, data_dir_fd);
	if (ret) {
		wprintw(recent_win, "Failed to save QSO: %s",
			xstrerror(ret));
		return ret; /* returning a negated errno */
	}

	ret = index_save(&current->qso);
	if (ret)
		wprintw(recent_win, "Warning: Failed to add QSO to index, "
			"UI may be out-of-sync: %s", xstrerror(ret));
	/* ignoring the index failure */

	/* display the uuid in the info message */
	snprintf(error, errorlen, "Saved %s", current->qso.uuid_str);

	current->dirty = 0;

	return CR_OK;
}

static enum cmd_result cmd_edit(char *arg, char *error, size_t errorlen)
{
	struct state *tab, *prev;
	struct xuuid uuid;
	int ret;

	if (!xuuid_parse(&uuid, arg))
		return -EINVAL;

	prev = current;

	/* open & activate a new tab */
	tab = tab_alloc();
	if (!tab)
		return CR_TOO_MANY_TABS;

	__activate_tab(tab);

	/* load entry from index */
	ret = index_get(&uuid, &current->qso);
	if (ret) {
		tab_close(current);
		__activate_tab(prev);
		return ret;
	}

	/* calculate derived values */
	recalculate_times();
	recalculate_location();
	recalculate_distance();

	/* everything is clean - we just loaded it */
	current->dirty = 0;

	return CR_OK;
}

static enum cmd_result __cmd_startend(char *arg, char *error, size_t errorlen,
				      bool start, enum refresh dirty)
{
	int ret;

	if (!strcasecmp(arg, "now"))
		ret = qso_set_datetime(&current->qso, start, NULL);
	else
		ret = qso_set_datetime(&current->qso, start, arg);

	if (ret) {
		snprintf(error, errorlen, "invalid date or time, specify "
			 "'now', a date, or a time");
		return ret;
	}

	recalculate_times();

	current->dirty |= dirty;

	return CR_OK;
}

static enum cmd_result cmd_start(char *arg, char *error, size_t errorlen)
{
	return __cmd_startend(arg, error, errorlen, true,
			      REF_DATE | REF_TIME_START);
}

static enum cmd_result cmd_end(char *arg, char *error, size_t errorlen)
{
	return __cmd_startend(arg, error, errorlen, false,
			      REF_DATE | REF_TIME_END);
}

static enum cmd_result cmd_qrz(char *arg, char *error, size_t errorlen)
{
	char url[256];
	char *cmd[] = {
		"firefox",
		"--new-tab",
		url,
		NULL,
	};
	int ret;

	if (!current->qso.rx.station_call) {
		snprintf(error, errorlen, "No callsign entered");
		return -EINVAL;
	}

	snprintf(url, sizeof(url), "https://qrz.com/db/%s",
		 str_cstr(current->qso.rx.station_call));

	ret = -posix_spawnp(NULL, cmd[0], NULL, NULL, cmd, NULL);
	if (ret)
		snprintf(error, errorlen, "Failed to start %s: %s", cmd[0],
			 xstrerror(ret));
	else
		snprintf(error, errorlen, "Opening qrz.com...");

	return ret;
}

static enum cmd_result cmd_de(char *arg, char *error, size_t errorlen)
{
	qso_set_call(&current->qso, true, arg);

	current->dirty |= REF_DE;

	return CR_OK;
}

static enum cmd_result cmd_de_grid(char *arg, char *error, size_t errorlen)
{
	qso_set_grid(&current->qso, true, arg);

	recalculate_distance();

	current->dirty |= REF_DE_GRID;

	return CR_OK;
}

static enum cmd_result cmd_de_power(char *arg, char *error, size_t errorlen)
{
	int ret;

	ret = qso_set_power_user(&current->qso, true, arg);
	if (ret) {
		snprintf(error, errorlen, "invalid TX power, specify "
			 "power in W without units (e.g., 6.7)");
		return ret;
	}

	current->dirty |= REF_DE_POWER;

	return CR_OK;
}

static enum cmd_result cmd_call(char *arg, char *error, size_t errorlen)
{
	struct country_result result;

	qso_set_call(&current->qso, false, arg);

	if (country_lookup(current->qso.rx.station_call, &result)) {
		qso_set_country(&current->qso, false, result.fullname);
		qso_set_zone_continent(&current->qso, false,
				       str_cstr(result.continent));
		qso_set_zone_dxcc(&current->qso, false, result.code);
		qso_set_zone_itu(&current->qso, false, result.itu);
		qso_set_zone_cqz(&current->qso, false, result.cqz);

		str_putref(result.continent);
	} else {
		qso_clear_country(&current->qso, false);
		qso_clear_zone_continent(&current->qso, false);
		qso_clear_zone_dxcc(&current->qso, false);
		qso_clear_zone_itu(&current->qso, false);
		qso_clear_zone_cqz(&current->qso, false);
	}

	recalculate_location();
	recalculate_distance();

	current->dirty |= REF_CALL;

	return CR_OK;
}

static enum cmd_result cmd_freq(char *arg, char *error, size_t errorlen)
{
	int ret;

	ret = qso_set_freq(&current->qso, true, arg);
	if (ret) {
		snprintf(error, errorlen, "invalid frequency, specify "
			 "frequency in MHz without units (e.g., 14.230)");
		return ret;
	}

	ret = qso_set_freq(&current->qso, false, arg);
	ASSERT0(ret); /* it worked on the tx side */

	current->dirty |= REF_FREQ;

	return CR_OK;
}

static enum cmd_result cmd_mode(char *arg, char *error, size_t errorlen)
{
	qso_set_mode(&current->qso, true, arg);
	qso_set_mode(&current->qso, false, arg);

	current->dirty |= REF_MODE;

	return CR_OK;
}

static enum cmd_result cmd_rst_s(char *arg, char *error, size_t errorlen)
{
	qso_set_rst(&current->qso, true, arg);

	current->dirty |= REF_RSTS;

	return CR_OK;
}

static enum cmd_result cmd_rst_r(char *arg, char *error, size_t errorlen)
{
	qso_set_rst(&current->qso, false, arg);

	current->dirty |= REF_RSTR;

	return CR_OK;
}

static enum cmd_result cmd_name(char *arg, char *error, size_t errorlen)
{
	qso_set_name(&current->qso, false, STR_DUP(arg));

	current->dirty |= REF_NAME;

	return CR_OK;
}

static enum cmd_result cmd_qth(char *arg, char *error, size_t errorlen)
{
	qso_set_qth(&current->qso, false, STR_DUP(arg));

	current->dirty |= REF_QTH;

	return CR_OK;
}

static enum cmd_result cmd_grid(char *arg, char *error, size_t errorlen)
{
	qso_set_grid(&current->qso, false, arg);

	recalculate_location();
	recalculate_distance();

	current->dirty |= REF_GRID;

	return CR_OK;
}

static enum cmd_result cmd_country(char *arg, char *error, size_t errorlen)
{
	qso_set_country(&current->qso, false, STR_DUP(arg));

	current->dirty |= REF_COUNTRY;

	return CR_OK;
}

static enum cmd_result cmd_state(char *arg, char *error, size_t errorlen)
{
	qso_set_state(&current->qso, false, STR_DUP(arg));

	current->dirty |= REF_STATE;

	return CR_OK;
}

static enum cmd_result cmd_county(char *arg, char *error, size_t errorlen)
{
	qso_set_county(&current->qso, false, STR_DUP(arg));

	current->dirty |= REF_COUNTY;

	return CR_OK;
}

static enum cmd_result cmd_continent(char *arg, char *error, size_t errorlen)
{
	qso_set_zone_continent(&current->qso, false, arg);

	current->dirty |= REF_ZONE_CONT;

	return CR_OK;
}

static enum cmd_result cmd_dxcc(char *arg, char *error, size_t errorlen)
{
	int ret;

	ret = qso_set_zone_dxcc_user(&current->qso, false, arg);
	switch (ret) {
		case 0:
			break;
		case -EINVAL:
		case -ERANGE:
			snprintf(error, errorlen, "invalid DXCC entity number");
			return ret;
		default:
			return ret;
	}

	current->dirty |= REF_ZONE_DXCC;

	return CR_OK;
}

static enum cmd_result cmd_itu(char *arg, char *error, size_t errorlen)
{
	int ret;

	ret = qso_set_zone_itu_user(&current->qso, false, arg);
	switch (ret) {
		case 0:
			break;
		case -EINVAL:
		case -ERANGE:
			snprintf(error, errorlen, "invalid ITU zone, valid "
				 "range is 1..90, 0 clears the field");
			return ret;
		default:
			return ret;
	}

	current->dirty |= REF_ZONE_ITU;

	return CR_OK;
}

static enum cmd_result cmd_cqz(char *arg, char *error, size_t errorlen)
{
	int ret;

	ret = qso_set_zone_cqz_user(&current->qso, false, arg);
	switch (ret) {
		case 0:
			break;
		case -EINVAL:
		case -ERANGE:
			snprintf(error, errorlen, "invalid CQ zone, valid "
				 "range is 1..40, 0 clears the field");
			return ret;
		default:
			return ret;
	}

	current->dirty |= REF_ZONE_CQZ;

	return CR_OK;
}

static enum cmd_result cmd_comment(char *arg, char *error, size_t errorlen)
{
	qso_set_comment(&current->qso, STR_DUP(arg));

	current->dirty |= REF_COMMENT;

	return CR_OK;
}

static enum cmd_result cmd_seq_r(char *arg, char *error, size_t errorlen)
{
	qso_set_seq_user(&current->qso, false, arg);

	current->dirty |= REF_SEQ;

	return CR_OK;
}

static enum cmd_result cmd_seq_s(char *arg, char *error, size_t errorlen)
{
	qso_set_seq_user(&current->qso, true, arg);

	current->dirty |= REF_SEQ;

	return CR_OK;
}

static void print_field(WINDOW *win, int row, int col, const char *str,
			size_t width, bool dirty)
{
	ssize_t len;

	wmove(win, row, col);

	if (!str)
		str = "";

	/*
	 * HACK: count number of printable characters
	 *
	 * We get UTF-8 strings passed in, but we need to know how wide it
	 * will print.  We assume that combining characters don't exist, and
	 * just count the number of bytes < 0x80 because that's how many
	 * codepoints there are.
	 */
	len = utf8_is_valid_str(str, strlen(str));
	ASSERT3S(len, >=, 0);

	if (len <= width) {
		size_t i;

		/* replace leading spaces with '_' */
		for (i = 0; (i < width) && (str[i] == ' '); i++)
			wprintw(win, "_");

		wattrset(win, dirty ? A_BOLD : A_NORMAL);
		wprintw(win, "%s", &str[i]);
		wattrset(win, A_NORMAL);

		/* padd with '_' */
		for (i = len; i < width; i++)
			wprintw(win, "_");
	} else {
		/* truncate & append ellipsis */
		wattrset(win, dirty ? A_BOLD : A_NORMAL);
		wprintw(win, "%.*s…", width - 1, str);
		wattrset(win, A_NORMAL);
	}
}

static int print_label(WINDOW *win, int row, int col, const char *str)
{
	bool emph;
	int count;
	size_t i;

	if (!str)
		return 0;

	wmove(win, row, col);

	for (i = 0, emph = false, count = 0; str[i]; i++) {
		if (str[i] == '_') {
			emph = true;
			continue;
		}

		waddch(win, str[i] | COLOR_PAIR(CMD_HINT_COLOR) |
		       (emph ? A_UNDERLINE : A_NORMAL));

		count++;
		emph = false;
	}

	return count;
}

static void print_tab_labels(void)
{
	char title[32];
	bool first;
	size_t i;

	/* TODO: We should just redraw the top line as needed */
	whrule(entry_win, 0, ENTRY_WIN_WIDTH);

	wmove(entry_win, 0, 2);

	wprintw(entry_win, "[");
	for (i = 0, first = true; i < ARRAY_LEN(states); i++) {
		const bool this = (current == &states[i]);

		if (!states[i].in_use)
			continue;

		if (states[i].qso.rx.station_call)
			strcpy_safe(title,
				    str_cstr(states[i].qso.rx.station_call),
				    sizeof(title));
		else
			strcpy(title, "???");

		if (!first)
			wprintw(entry_win, " ");

		wattrset(entry_win, this ? A_REVERSE : A_NORMAL);
		wprintw(entry_win, "%zu:%s", i, title);
		wattrset(entry_win, A_NORMAL);

		first = false;
	}
	wprintw(entry_win, "]");
}

static void setup_windows(void)
{
	size_t i;

	entry_win = newwin(ENTRY_WIN_HEIGHT, ENTRY_WIN_WIDTH, ENTRY_WIN_ROW, 0);
	whrule(entry_win, 0, ENTRY_WIN_WIDTH);

	for (i = 0; i < ARRAY_LEN(entry_fields); i++) {
		const struct entry_field *entry = &entry_fields[i];

		print_label(entry_win, entry->row + 1, entry->col_label,
			    entry->label);
	}

	/* TODO: print '-' to separate the times */

	print_tab_labels();

	call_win = newwin(CALL_WIN_HEIGHT, CALL_WIN_WIDTH, CALL_WIN_ROW, 0);

	recent_win = newwin(RECENT_WIN_HEIGHT, RECENT_WIN_WIDTH, RECENT_WIN_ROW, 0);
	whrule(recent_win, 0, RECENT_WIN_WIDTH);

	cmd_win = newwin(CMD_WIN_HEIGHT, CMD_WIN_WIDTH, CMD_WIN_ROW, 0);
	whrule(cmd_win, 0, CMD_WIN_WIDTH);

	mvwprintw(cmd_win, 1, 0, "> ");
}

static void refresh_windows(void)
{
	wrefresh(entry_win);
	wrefresh(call_win);
	wrefresh(recent_win);
	wrefresh(cmd_win);
}

/* pretty-print the power with appropriate units */
static bool format_power(char *buf, size_t bufsize, struct val *power)
{
	uint64_t v = power ? power->i : 0;
	const char *unit;
	int width;

	if (!v) {
		buf[0] = '\0';
		return false; /* no power */
	}

	if (v < 1000) {
		unit = "mW";
		width = 3;
	} else {
		unit = "W";
		width = 4;

		/* round up */
		if ((v % 1000) >= 500)
			v += 1000;

		v /= 1000;
	}

	snprintf(buf, bufsize, "%*"PRIu64"%s", width, v, unit);

	return true;
}

/* pretty-print the frequency, omitting any trailing zeros */
static bool format_freq(char *buf, size_t bufsize, struct val *freq)
{
	const uint64_t v = freq ? freq->i : 0;
	uint64_t mhz = v / 1000000;
	uint64_t hz = v % 1000000;
	int digits = 6; /* for Hz */

	if (!v) {
		buf[0] = '\0';
		return false; /* no freq */
	}

	/*
	 * MHz can be used as-is, but we want to omit as many zeros after
	 * the decimal point as possible.  Therefore, we tweak the Hz value
	 * and rely to the following snprintf to make them reasonable
	 * looking.
	 */

	while (hz && ((hz % 10) == 0)) {
		hz /= 10;
		digits--;
	}

	snprintf(buf, bufsize, "%"PRIu64".%0*"PRIu64, mhz, digits, hz);

	return true;
}

static void print_qso_line(WINDOW *win, struct qso *qso)
{
	char tmp[32];

	/* YYYY-MM-DD HH:SS WX1YZ_____ 144.444___ FM___ 599 599 10W FN42 */
	time_snprint_date(tmp, sizeof(tmp), &qso->start);
	wprintw_cstr(win, tmp, 10); /* date */
	time_snprint_time(tmp, sizeof(tmp), &qso->start);
	wprintw_cstr(win, tmp, 5); /* time */
	wprintw_cstr(win, str_cstr(qso->rx.station_call), 10); /* call */
	format_freq(tmp, sizeof(tmp), qso->tx.freq);
	wprintw_cstr(win, tmp, 10); /* freq */
	wprintw_cstr(win, str_cstr(qso->tx.mode), 6); /* mode */
	wprintw_cstr(win, str_cstr(qso->tx.rst), 3); /* RST */
	wprintw_cstr(win, str_cstr(qso->rx.rst), 3); /* RST */
	format_power(tmp, sizeof(tmp), qso->tx.power);
	wprintw_cstr(win, tmp, 5); /* power */
	wprintw_cstr(win, str_cstr(qso->rx.grid), 8); /* grid */
	/* TODO: print flags */
}

static size_t __get_band(uint64_t band)
{
	size_t i;

	for (i = 0; i < ARRAY_LEN(summary_bands); i++)
		if (summary_bands[i] == band)
			return i;

	/* default to unknown/other column */
	ASSERT(!summary_bands[0]);

	return 0;
}

static size_t __get_mode(struct stats_summary *summary, struct str *mode)
{
	size_t min;
	size_t i;

	/* find matching mode name */
	for (i = 0; i < ARRAY_LEN(summary->mode); i++) {
		if (!summary->mode[i].total)
			continue;

		if ((summary->mode[i].mode_name && !mode) ||
		    (!summary->mode[i].mode_name && mode))
			continue;

		if (mode && str_cmp(summary->mode[i].mode_name, mode))
			continue;

		return i;
	}

	/* find an unused spot, and use it */
	for (i = 0; i < ARRAY_LEN(summary->mode); i++) {
		if (summary->mode[i].total)
			continue;

		summary->mode[i].mode_name = mode;

		return i;
	}

	/* find unknown/other and use it */
	for (i = 0; i < ARRAY_LEN(summary->mode); i++) {
		ASSERT(summary->mode[i].total);

		if (summary->mode[i].mode_name)
			continue;

		return i;
	}

	/* find least used mode, and convert it to unknown/other */
	for (i = 1, min = 0; i < ARRAY_LEN(summary->mode); i++) {
		ASSERT(summary->mode[i].total);
		ASSERT(summary->mode[i].mode_name);

		if (summary->mode[i].total < summary->mode[min].total)
			min = i;
	}

	summary->mode[min].mode_name = NULL;

	return min;
}

static void summarize_stats(struct stats_summary *summary)
{
	struct index_iter_ctx ctx;
	struct index_stat *stat;
	size_t mode, band;

	memset(summary, 0, sizeof(*summary));

	for (stat = index_iter_call_stats_start(current->qso.rx.station_call, &ctx);
	     stat;
	     stat = index_iter_call_stats_next(&ctx)) {
		band = __get_band(stat->band);
		mode = __get_mode(summary, stat->mode);

		summary->mode[mode].total += stat->count;
		summary->mode[mode].band[band] += stat->count;
	}
}

static void refresh_fields(enum refresh refresh)
{
	size_t i;

	for (i = 0; i < ARRAY_LEN(entry_fields); i++) {
		const struct entry_field *entry = &entry_fields[i];
		uintptr_t ptr;

		if (!(entry->refresh & refresh))
			continue;

		ptr = ((uintptr_t) current) + entry->offset;

		switch (entry->type) {
			case VT_STR:
				print_field(entry_win, entry->row + 1,
					    entry->col_value,
					    str_cstr(*(struct str **) ptr),
					    entry->len,
					    !!(current->dirty & entry->refresh));
				break;
			case VT_INT: {
				struct val *v = *(struct val **) ptr;
				char buf[22];

				/* default to the empty string */
				buf[0] = '\0';

				switch (entry->refresh) {
					case REF_DE_POWER:
						format_power(buf, sizeof(buf), v);
						break;
					case REF_FREQ:
						format_freq(buf, sizeof(buf), v);
						break;
					case REF_ZONE_ITU:
					case REF_ZONE_CQZ:
						if (v)
							snprintf(buf, sizeof(buf),
								 "%02"PRIu64, v->i);
						break;
					case REF_ZONE_DXCC:
						if (v)
							snprintf(buf, sizeof(buf),
								 "%03"PRIu64, v->i);
						break;
					case REF_DE:
					case REF_DE_GRID:
					case REF_DATE:
					case REF_TIME_START:
					case REF_TIME_END:
					case REF_CALL:
					case REF_MODE:
					case REF_RSTS:
					case REF_RSTR:
					case REF_NAME:
					case REF_QTH:
					case REF_GRID:
					case REF_COUNTRY:
					case REF_GEO:
					case REF_STATE:
					case REF_COUNTY:
					case REF_ZONE_CONT:
					case REF_COMMENT:
					case REF_CALL_WIN:
					case REF_RECENT_WIN:
					case REF_SEQ:
						panic("impossible");
				}

				print_field(entry_win, entry->row + 1,
					    entry->col_value,
					    buf, entry->len,
					    !!(current->dirty & entry->refresh));
				break;
			}
			default:
				panic("unhandled entry value type: %d", entry->type);
		}
	}

	if (refresh & REF_CALL_WIN) {
		/* index -> band in mm */
		struct stats_summary summary;
		struct index_iter_ctx ctx;
		struct qso *qso;
		size_t i, j;
		int row;

		/* clear it */
		for (row = 1; row < CALL_WIN_HEIGHT; row++)
		     mvwprintw(call_win, row, 0, "%*s", CALL_WIN_WIDTH - 1, "");

		summarize_stats(&summary);

		/* print header */
		mvwprintw(call_win, 1, 7, " #");
		for (i = 0; i < ARRAY_LEN(summary_bands); i++) {
			const int col = 10 + i * 4;

			if (!summary_bands[i])
				mvwprintw(call_win, 1, col, "U/O");
			else if (summary_bands[i] > 1000)
				mvwprintw(call_win, 1, col, "%3.0f",
					  summary_bands[i] / 1000.);
			else
				mvwprintw(call_win, 1, col, "%3.1f",
					  summary_bands[i] / 1000.);
		}

		/* print summary */
		row = 2;
		for (i = 0; i < ARRAY_LEN(summary.mode); i++) {
			const char *mode_name;

			if (!summary.mode[i].total)
				continue;

			/* print the mode name */
			if (summary.mode[i].mode_name)
				mode_name = str_cstr(summary.mode[i].mode_name);
			else
				mode_name = "<U/O>";

			wmove(call_win, row, 0);
			wprintw_cstr(call_win, mode_name, 6);

			/* print the mode total count */
			if (summary.mode[i].total < 100)
				wprintw(call_win, "%2"PRIu64,
					summary.mode[i].total);
			else
				wprintw(call_win, ">>");

			/* print per band indicator */
			for (j = 0; j < ARRAY_LEN(summary.mode[i].band); j++) {
				const int col = 10 + j * 4 + 1;

				if (!summary.mode[i].band[j])
					continue;

				mvwprintw(call_win, row, col, "X");
			}

			row++;
		}

		row++; /* skip a row */

		/* for each contact, print a line */
		for (qso = index_iter_call_history_start(current->qso.rx.station_call, &ctx);
		     qso;
		     qso = index_iter_call_history_next(&ctx), row++) {
			wmove(call_win, row, 0);
			print_qso_line(call_win, qso);

			if (row >= (CALL_WIN_HEIGHT - 1))
				break;
		}
	}

	if (refresh & REF_RECENT_WIN) {
		struct index_iter_ctx ctx;
		struct qso *qso;
		int row;

		/* clear it */
		for (row = 1; row < RECENT_WIN_HEIGHT; row++)
		     mvwprintw(recent_win, row, 0, "%*s", RECENT_WIN_WIDTH - 1, "");

		/* for each mode, print a line */
		for (qso = index_iter_history_start(&ctx, false), row = 1;
		     qso;
		     qso = index_iter_history_next(&ctx), row++) {
			wmove(recent_win, row, 0);
			print_qso_line(recent_win, qso);

			if (row >= (RECENT_WIN_HEIGHT - 1))
				break;
		}
	}
}

static enum cmd_result parse_command(char *buf, char *error, size_t errorlen)
{
	char *cmdname;
	size_t i;

	/* strip whitespace from buf - the command name */
	buf = trim_ws(buf);

	cmdname = strsep(&buf, " \t");

	/* strip whitespace from buf - the argument */
	buf = trim_ws(buf);

	for (i = 0; i < ARRAY_LEN(commands); i++) {
		const struct command *cmd = &commands[i];
		enum cmd_result res;

		if (strcasecmp(cmdname, cmd->cmd))
			continue;

		if (cmd->arg_required && !(buf && strlen(buf))) {
			snprintf(error, errorlen, "command '%s' requires an "
				 "argument", cmd->cmd);
			return CR_MISSING_ARG;
		}
		if (!cmd->arg_required && (buf && strlen(buf))) {
			snprintf(error, errorlen, "command '%s' doesn't take "
				 "arguments", cmd->cmd);
			return CR_EXTRA_ARG;
		}

		res = cmd->fxn(buf, error, errorlen);
		if (res)
			return res;

		refresh_fields(cmd->refresh);
		return CR_OK;
	}

	return CR_INVAL_CMD;
}

static void usage(const char *prog)
{
	fprintf(stderr, "HLog version %s (libjeffpc %s)\n\n", version_string,
		jeffpc_version);

	fprintf(stderr, "Usage: %s <path to data dir>\n", prog);
	exit(1);
}

static int load_bands(void)
{
	int ret;

	printf("Loading band data...\n");
	ret = bands_load(data_dir_fd, DATA_DIR_BAND_FNAME);
	if (ret == -ENOENT)
		ret = bands_load(AT_FDCWD, DATA_DIR_BAND_FNAME);
	if (ret == -ENOENT)
		ret = bands_load(AT_FDCWD,
				 INSTALLED_DATA_DIR "/" DATA_DIR_BAND_FNAME);
	printf("%s.\n", ret ? "error" : "done");

	if (ret)
		fprintf(stderr, "Error: failed to load band data: %s\n",
			xstrerror(ret));

	return ret;
}

static int load_countries(void)
{
	int ret;

	printf("Loading countries data...\n");
	ret = country_load(data_dir_fd, DATA_DIR_COUNTRY_FNAME);
	printf("%s.\n", ret ? "error" : "done");

	if (ret)
		fprintf(stderr, "Error: failed to load country data: %s\n",
			xstrerror(ret));

	return ret;
}

static int load_index(void)
{
	int ret;

	printf("Loading existing QSOs...\n");
	ret = index_init(data_dir_fd);
	printf("%s.\n", ret ? "error" : "done");

	if (ret)
		fprintf(stderr, "Error: failed to load existing entries: %s\n",
			xstrerror(ret));

	return ret;
}

static void state_init(void)
{
	size_t i;

	/* reset the state */
	for (i = 0; i < ARRAY_LEN(states); i++) {
		qso_reset(&states[i].qso, true);

		states[i].date = &default_date;
		states[i].time_start = &default_time;
		states[i].time_end = &default_time;
		states[i].latlon = &default_latlon;
		states[i].azidist = &default_azidist;

		states[i].in_use = false;
	}

	/* we always start with the first one */
	current = &states[0];
	current->in_use = true;
}

static int fill_str(struct val *parsed, const char *name,
		    char *buf, size_t buflen)
{
	struct str *str;
	int ret;

	str = sexpr_alist_lookup_str(parsed, name);
	if (!str)
		return -ENOENT;

	ret = strcpy_safe(buf, str_cstr(str), buflen);

	str_putref(str);

	return ret;
}

#define LOAD_CONFIG_STR(parsed, name, tmp, setter) \
	do { \
		int ret; \
		ret = fill_str((parsed), (name), (tmp), sizeof(tmp)); \
		if (!ret) { \
			for (size_t i = 0; i < ARRAY_LEN(states); i++) \
				setter; \
		} \
	} while (0)

static int load_config(void)
{
	char tmp[128];
	struct val *parsed;
	const char *raw;
	uint64_t uint;
	size_t len;
	size_t i;
	int ret;

	raw = read_file_len_at(data_dir_fd, DATA_DIR_CONFIG_FNAME, &len);
	if (IS_ERR(raw))
		return PTR_ERR(raw);

	parsed = sexpr_parse(raw, len);
	if (IS_ERR(parsed)) {
		ret = PTR_ERR(parsed);
		goto err;
	}

	/* extract the values (NULL strings and 0 ints are ok) */
	LOAD_CONFIG_STR(parsed, "call", tmp,
			qso_set_call(&states[i].qso, true, tmp));
	LOAD_CONFIG_STR(parsed, "grid", tmp,
			qso_set_grid(&states[i].qso, true, tmp));
	LOAD_CONFIG_STR(parsed, "mode", tmp,
			qso_set_mode(&states[i].qso, true, tmp));
	LOAD_CONFIG_STR(parsed, "mode", tmp,
			qso_set_mode(&states[i].qso, false, tmp));

	uint = sexpr_alist_lookup_int(parsed, "power", NULL);
	for (i = 0; i < ARRAY_LEN(states); i++)
		qso_set_power(&states[i].qso, true, uint);

	val_putref(parsed);

	ret = 0;

err:
	free((void *) raw);

	if (ret)
		fprintf(stderr, "Error: failed to load config: %s\n",
			xstrerror(ret));

	return ret;
}

static void fill_generic_error(char *buf, size_t bufsize, int res)
{
#define COPYSTR(v) strcpy_safe(buf, (v), bufsize);
	if (res < 0) {
		COPYSTR(xstrerror(res));
	} else {
		switch ((enum cmd_result) res) {
			case CR_OK:
				panic("impossible");
			case CR_INVAL_CMD:
				COPYSTR("invalid command");
				break;
			case CR_MISSING_ARG:
				COPYSTR("command requires an argument");
				break;
			case CR_EXTRA_ARG:
				COPYSTR("too many arguments");
				break;
			case CR_TOO_MANY_TABS:
				snprintf(buf, bufsize, "cannot open more than "
					 "%d tabs", MAX_STATE_COUNT);
				break;
			case CR_INVAL_TAB_ID:
				COPYSTR("invalid tab ID");
				break;
			case CR_UNKNOWN_TAB_ID:
				COPYSTR("tab doesn't exist");
				break;
		}
	}
#undef COPYSTR
}

int main(int argc, char **argv)
{
	int ret;

	setlocale(LC_ALL, "");
	ASSERT0(putenv("TZ=UTC"));

	if (argc != 2)
		usage(argv[0]);

	data_dir = argv[1];
	data_dir_fd = xopen(data_dir, O_RDONLY, 0);
	if (data_dir_fd < 0) {
		fprintf(stderr, "Error: failed to open data directory: %s\n",
			xstrerror(data_dir_fd));
		goto err;
	}

	ret = load_bands();
	if (ret)
		goto err_close;

	ret = load_countries();
	if (ret)
		goto err_close;

	ret = load_index();
	if (ret)
		goto err_close;

	state_init();

	ret = load_config();
	if (ret)
		goto err_close;

	initscr();
	start_color();

	init_pair(CMD_HINT_COLOR, COLOR_YELLOW, COLOR_BLACK);

	setup_windows();
	refresh_fields(~0u);
	refresh_windows();

	for (;;) {
		char buf[CMD_WIN_WIDTH - 2 /* border */ - 2 /* prompt */ - 1];

		/* TODO: add line editing support */

		if (wgetnstr(cmd_win, buf, sizeof(buf)) != ERR) {
			char error[CMD_WIN_WIDTH];
			enum cmd_result res;

			error[0] = '\0'; /* empty the error string */

			res = parse_command(buf, error, sizeof(error));
			if (res != CR_OK) {
				if (!error[0])
					fill_generic_error(error, sizeof(error),
							   res);

				mvwprintw(cmd_win, 2, 0, "Error: %s", error);
				beep();
			} else {
				/* print info message / clear previous error */
				mvwprintw(cmd_win, 2, 0, "%-*s", CMD_WIN_WIDTH,
					  error);
			}
		}

		/* clear the previous command */
		mvwprintw(cmd_win, 1, 2, "%*s", CMD_WIN_WIDTH - 3, "");

		print_tab_labels();

		wmove(cmd_win, 1, 2);
		refresh_windows();
	}

	delwin(entry_win);
	delwin(call_win);
	delwin(recent_win);

	endwin();

	xclose(data_dir_fd);

	return 0;

err_close:
	xclose(data_dir_fd);

err:
	return 1;
}
