/*
 * Copyright (c) 2020 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <jeffpc/types.h>

#include <ctype.h>
#include <string.h>

#include "maidenhead.h"

static inline bool decode_thing(double *loc, char start, char end,
				double degrees, double *last_step, char c)
{
	c = toupper(c);

	if ((c < start) || (c > end))
		return false;

	*loc += (c - start) * degrees;
	*last_step = degrees;

	return true;
}

bool maidenhead_decode(const char *grid, double *lat_r, double *lon_r)
{
	double lat, lon;
	double lat_step, lon_step;
	size_t len;

	if (!grid || (strlen(grid) < 2))
		return false;

	len = strlen(grid);
	lat = -90.;
	lon = -180.;

	/* decode the field */
	if (!decode_thing(&lon, 'A', 'R', 20., &lon_step, grid[0]))
		return false;
	if (!decode_thing(&lat, 'A', 'R', 10., &lat_step, grid[1]))
		return false;

	/* decode the square */
	if ((len >= 3) &&
	    !decode_thing(&lon, '0', '9', 2., &lon_step, grid[2]))
		return false;
	if ((len >= 4) &&
	    !decode_thing(&lat, '0', '9', 1., &lat_step, grid[3]))
		return false;

	/* decode the subsquare */
	if ((len >= 5) &&
	    !decode_thing(&lon, 'A', 'X', 5./60., &lon_step, grid[4]))
		return false;
	if ((len >= 6) &&
	    !decode_thing(&lat, 'A', 'X', 2.5/60., &lat_step, grid[5]))
		return false;

	/* decode the extended square */
	if ((len >= 7) &&
	    !decode_thing(&lon, '0', '9', 5./600., &lon_step, grid[6]))
		return false;
	if ((len >= 8) &&
	    !decode_thing(&lat, '0', '9', 2.5/600., &lat_step, grid[7]))
		return false;

	*lat_r = lat + lat_step / 2;
	*lon_r = lon + lon_step / 2;

	return true;
}
