/*
 * Copyright (c) 2020 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <ctype.h>
#include <time.h>

#include <jeffpc/buffer.h>
#include <jeffpc/io.h>
#include <jeffpc/time.h>
#include <jeffpc/rand.h>

#include "qso.h"
#include "qso_impl.h"
#include "adif.h"
#include "util.h"

int qso_save(const struct qso *qso, int data_dir)
{
	char tmpname[FILENAME_MAX];
	char qsoname[FILENAME_MAX];
	struct buffer *buf;
	int ret;

	buf = qso_pack_sexpr(qso);
	if (IS_ERR(buf))
		return PTR_ERR(buf);

	snprintf(tmpname, sizeof(tmpname), "tmp/%s.scm", qso->uuid_str);
	snprintf(qsoname, sizeof(qsoname), "qso/%s.scm", qso->uuid_str);

	ret = xunlinkat(data_dir, tmpname, 0);
	if (ret && (ret != -ENOENT))
		goto err;

	ret = write_file_at(data_dir, tmpname, buffer_data(buf),
			    buffer_size(buf));
	if (ret)
		goto err;

	/* TODO: sync the file */

	ret = xrenameat(data_dir, tmpname, data_dir, qsoname);

	/* TODO: fsync the dir */

err:
	buffer_free(buf);

	return ret;
}

static void qso_side_reset(struct qso *qso, bool full, bool local)
{
	if (full) {
		qso_clear_call(qso, local);
		if (local) {
			str_putref(qso->tx.operator_call);
			qso->tx.operator_call = NULL;
		} else {
			str_putref(qso->rx.operator_call);
			qso->rx.operator_call = NULL;
		}
		qso_clear_power(qso, local);
	}
	qso_clear_freq(qso, local);
	qso_clear_mode(qso, local);
	qso_clear_rst(qso, local);
	qso_clear_name(qso, local);
	qso_clear_qth(qso, local);
	if (full)
		qso_clear_grid(qso, local);

	qso_clear_country(qso, local);
	qso_clear_state(qso, local);
	qso_clear_county(qso, local);
	qso_clear_zone_continent(qso, local);
	qso_clear_zone_dxcc(qso, local);
	qso_clear_zone_itu(qso, local);
	qso_clear_zone_cqz(qso, local);
	qso_clear_seq(qso, local);
}

void qso_reset(struct qso *qso, bool full)
{
	qso_clear_datetime(qso, true);
	qso_clear_datetime(qso, false);
	qso_clear_comment(qso);

	qso_side_reset(qso, full, true);
	qso_side_reset(qso, true, false);

	nvl_putref(qso->additional);
	qso->additional = nvl_alloc();
	ASSERT(!IS_ERR(qso->additional));

	xuuid_generate(&qso->uuid);
	xuuid_unparse(&qso->uuid, qso->uuid_str);
}

static void qso_copy_side(struct qso_side *dst, struct qso_side *src)
{
	dst->station_call = str_getref(src->station_call);
	dst->operator_call = str_getref(src->operator_call);
	dst->power = val_getref(src->power);
	dst->freq = val_getref(src->freq);
	dst->mode = str_getref(src->mode);
	dst->rst = str_getref(src->rst);
	dst->name = str_getref(src->name);
	dst->qth = str_getref(src->qth);
	dst->grid = str_getref(src->grid);
	dst->country = str_getref(src->country);
	dst->state = str_getref(src->state);
	dst->county = str_getref(src->county);
	dst->continent = str_getref(src->continent);
	dst->dxcc = val_getref(src->dxcc);
	dst->itu = val_getref(src->itu);
	dst->cqz = val_getref(src->cqz);
	dst->seq = val_getref(src->seq);
}

void qso_copy(struct qso *dst, struct qso *src)
{
	qso_reset(dst, true);

	dst->uuid = src->uuid;
	memcpy(dst->uuid_str, src->uuid_str, sizeof(src->uuid_str));
	dst->start = src->start;
	dst->end = src->end;
	dst->comment = str_getref(src->comment);

	qso_copy_side(&dst->tx, &src->tx);
	qso_copy_side(&dst->rx, &src->rx);

	nvl_putref(dst->additional);
	dst->additional = nvl_getref(src->additional);
}

void qso_set_call(struct qso *qso, bool local, char *call)
{
	struct qso_side *side = local ? &qso->tx : &qso->rx;

	uppercase_inplace(call);

	str_putref(side->station_call);
	side->station_call = STR_DUP(call);
}

void qso_clear_call(struct qso *qso, bool local)
{
	struct qso_side *side = local ? &qso->tx : &qso->rx;

	str_putref(side->station_call);
	side->station_call = NULL;
}

static void qso_set_operator(struct qso *qso, bool local, char *call)
{
	struct qso_side *side = local ? &qso->tx : &qso->rx;

	uppercase_inplace(call);

	str_putref(side->operator_call);
	side->operator_call = STR_DUP(call);
}

int qso_set_freq(struct qso *qso, bool local, char *str)
{
	struct qso_side *side = local ? &qso->tx : &qso->rx;
	int64_t freq;
	int ret;

	ret = extract_float_as_int64(str, 1000000, &freq);
	if (ret)
		return ret;

	val_putref(side->freq);

	if (!freq)
		side->freq = NULL;
	else
		side->freq = VAL_ALLOC_INT(freq);

	return 0;
}

void qso_clear_freq(struct qso *qso, bool local)
{
	struct qso_side *side = local ? &qso->tx : &qso->rx;

	val_putref(side->freq);
	side->freq = NULL;
}

void qso_set_grid(struct qso *qso, bool local, char *grid)
{
	struct qso_side *side = local ? &qso->tx : &qso->rx;

	/* fix the case to match 'GO11nl22' */
	lowercase_inplace(grid);
	grid[0] = toupper(grid[0]);
	grid[1] = toupper(grid[1]);

	str_putref(side->grid);
	side->grid = STR_DUP(grid);
}

void qso_clear_grid(struct qso *qso, bool local)
{
	struct qso_side *side = local ? &qso->tx : &qso->rx;

	str_putref(side->grid);
	side->grid = NULL;
}

void qso_set_power(struct qso *qso, bool local, uint64_t milliwatts)
{
	struct qso_side *side = local ? &qso->tx : &qso->rx;

	val_putref(side->power);

	if (!milliwatts)
		side->power = NULL;
	else
		side->power = VAL_ALLOC_INT(milliwatts);
}

int qso_set_power_user(struct qso *qso, bool local, char *power)
{
	int64_t milliwatts;
	int ret;

	ret = extract_float_as_int64(power, 1000, &milliwatts);
	if (ret)
		return ret;

	qso_set_power(qso, local, milliwatts);

	return 0;
}

void qso_clear_power(struct qso *qso, bool local)
{
	struct qso_side *side = local ? &qso->tx : &qso->rx;

	val_putref(side->power);
	side->power = NULL;
}

/*
 * Try to parse using different date & time formats.
 *
 * We make up to two passes through the list of formats.  If the first pass
 * resulted in successful parsing, we advance the input past the parsed
 * values and try again.  The goal is to allow parsing of input with both
 * date and time (e.g., "2020-07-10 02:43:28").
 */
static int __qso_set_datetime(struct time *time, char *str)
{
	static struct {
		const char *fmt;
		enum time_parts have;
	} formats[] = {
		{ "%Y-%m-%d",	TIME_PART_DATE },
		{ "%Y%m%d",	TIME_PART_DATE },
		{ "%H:%M:%SZ",	TIME_PART_TIME_HMS },
		{ "%H:%M:%S",	TIME_PART_TIME_HMS },
		{ "%H:%MZ",	TIME_PART_TIME_HM },
		{ "%H:%M",	TIME_PART_TIME_HM },
		{ "%H%M%SZ",	TIME_PART_TIME_HMS },
		{ "%H%M%S",	TIME_PART_TIME_HMS },
		{ "%H%MZ",	TIME_PART_TIME_HM },
		{ "%H%M",	TIME_PART_TIME_HM },
	};
	bool first_pass;
	size_t i;

	memset(time, 0, sizeof(struct time));

	first_pass = true;

restart:
	for (i = 0; i < ARRAY_LEN(formats); i++) {
		struct tm tm;
		char *next;

		next = strptime(str, formats[i].fmt, &tm);
		if (next == NULL)
			continue;

		/* what is after the parsed chars? */
		switch (*next) {
			case ' ':
			case '\t':
			case '\0':
				/* end or whitespace - ok */
				next = find_nonws(next);
				break;
			default:
				/* garbage at the end - ignore format */
				continue;
		}

		/* parsed ok, but is it valid? */
		if ((formats[i].have & TIME_PART_DATE) &&
		    !valid_date(tm.tm_year + 1900, tm.tm_mon + 1, tm.tm_mday))
			continue; /* impossible date */
		if ((formats[i].have & TIME_PART_TIME_HM) &&
		    ((tm.tm_hour > 23) || (tm.tm_min > 59)))
			continue; /* impossible hours/minutes */
		if ((formats[i].have & TIME_PART_TIME_S) && (tm.tm_sec > 60))
			continue; /* impossible seconds */

		/* did we get two times or two dates? */
		if (formats[i].have & time->have)
			return -EINVAL;

		if (formats[i].have & TIME_PART_DATE) {
			time->year = tm.tm_year + 1900;
			time->month = tm.tm_mon + 1;
			time->day = tm.tm_mday;
			time->have |= TIME_PART_DATE;
		}

		if (formats[i].have & TIME_PART_TIME_HM) {
			time->hour = tm.tm_hour;
			time->min = tm.tm_min;
			time->have &= ~TIME_PART_TIME_S; /* clear the seconds */
			time->have |= TIME_PART_TIME_HM;
		}

		if (formats[i].have & TIME_PART_TIME_S) {
			time->sec = tm.tm_sec;
			time->have |= TIME_PART_TIME_S;
		}

		/* if there is more to parse, do a second pass */
		if (first_pass && *next) {
			str = next;
			first_pass = false;
			goto restart;
		}

		return 0;
	}

	return -EINVAL;
}

int qso_set_datetime(struct qso *qso, bool start, char *str)
{
	struct time *ts = start ? &qso->start : &qso->end;
	struct time tmp;
	int ret;

	if (!str) {
		/* set it to the current date & time */
		struct tm tm;
		time_t now;

		now = time(NULL);
		gmtime_r(&now, &tm);

		ts->year = tm.tm_year + 1900;
		ts->month = tm.tm_mon + 1;
		ts->day = tm.tm_mday;
		ts->hour = tm.tm_hour;
		ts->min = tm.tm_min;
		ts->sec = tm.tm_sec;

		ts->have = TIME_PART_DATE | TIME_PART_TIME_HMS;

		return 0;
	}

	ret = __qso_set_datetime(&tmp, str);
	if (ret)
		return ret;

	/* fill in qso time */

	if (tmp.have & TIME_PART_DATE) {
		ts->year = tmp.year;
		ts->month = tmp.month;
		ts->day = tmp.day;
		ts->have |= TIME_PART_DATE;
	}

	if (tmp.have & TIME_PART_TIME_HM) {
		ts->hour = tmp.hour;
		ts->min = tmp.min;
		ts->have &= ~TIME_PART_TIME_S; /* clear the seconds */
		ts->have |= TIME_PART_TIME_HM;
	}

	if (tmp.have & TIME_PART_TIME_S) {
		ts->sec = tmp.sec;
		ts->have |= TIME_PART_TIME_S;
	}

	return 0;
}

void qso_clear_datetime(struct qso *qso, bool start)
{
	struct time *ts = start ? &qso->start : &qso->end;

	ts->have = 0;
}

void qso_set_mode(struct qso *qso, bool local, char *mode)
{
	static struct str common_modes[] = {
		STR_STATIC_INITIALIZER("CW"),
		STR_STATIC_INITIALIZER("AM"),
		STR_STATIC_INITIALIZER("FM"),
		STR_STATIC_INITIALIZER("SSB"),
		STR_STATIC_INITIALIZER("USB"),
		STR_STATIC_INITIALIZER("LSB"),
		STR_STATIC_INITIALIZER("RTTY"),
		STR_STATIC_INITIALIZER("FT4"),
		STR_STATIC_INITIALIZER("FT8"),
		STR_STATIC_INITIALIZER("PSK31"),
		STR_STATIC_INITIALIZER("QPSK31"),
		STR_STATIC_INITIALIZER("OLIVIA"),
		STR_STATIC_INITIALIZER("DOMEX8"),
		STR_STATIC_INITIALIZER("THOR4"),
	};
	struct qso_side *side = local ? &qso->tx : &qso->rx;
	size_t i;

	if (side->mode &&
	    !strcmp(str_cstr(side->mode), mode))
		return; /* no change */

	str_putref(side->mode);

	for (i = 0; i < ARRAY_LEN(common_modes); i++) {
		if (strcasecmp(str_cstr(&common_modes[i]), mode))
			continue;

		side->mode = &common_modes[i];
		return;
	}

	/* non-common mode, use it as-is */
	side->mode = STR_DUP(mode);
}

void qso_clear_mode(struct qso *qso, bool local)
{
	struct qso_side *side = local ? &qso->tx : &qso->rx;

	str_putref(side->mode);
	side->mode = NULL;
}

void qso_set_rst(struct qso *qso, bool local, char *rst)
{
	struct qso_side *side = local ? &qso->tx : &qso->rx;

	str_putref(side->rst);
	side->rst = STR_DUP(rst);
}

void qso_clear_rst(struct qso *qso, bool local)
{
	struct qso_side *side = local ? &qso->tx : &qso->rx;

	str_putref(side->rst);
	side->rst = NULL;
}

void qso_set_name(struct qso *qso, bool local, struct str *name)
{
	struct qso_side *side = local ? &qso->tx : &qso->rx;

	str_putref(side->name);
	side->name = name;
}

void qso_clear_name(struct qso *qso, bool local)
{
	struct qso_side *side = local ? &qso->tx : &qso->rx;

	str_putref(side->name);
	side->name = NULL;
}

void qso_set_qth(struct qso *qso, bool local, struct str *qth)
{
	struct qso_side *side = local ? &qso->tx : &qso->rx;

	str_putref(side->qth);
	side->qth = qth;
}

void qso_clear_qth(struct qso *qso, bool local)
{
	struct qso_side *side = local ? &qso->tx : &qso->rx;

	str_putref(side->qth);
	side->qth = NULL;
}

void qso_set_country(struct qso *qso, bool local, struct str *country)
{
	struct qso_side *side = local ? &qso->tx : &qso->rx;

	str_putref(side->country);
	side->country = country;
}

void qso_clear_country(struct qso *qso, bool local)
{
	struct qso_side *side = local ? &qso->tx : &qso->rx;

	str_putref(side->country);
	side->country = NULL;
}

void qso_set_state(struct qso *qso, bool local, struct str *state)
{
	struct qso_side *side = local ? &qso->tx : &qso->rx;

	str_putref(side->state);
	side->state = state;
}

void qso_clear_state(struct qso *qso, bool local)
{
	struct qso_side *side = local ? &qso->tx : &qso->rx;

	str_putref(side->state);
	side->state = NULL;
}

void qso_set_county(struct qso *qso, bool local, struct str *county)
{
	struct qso_side *side = local ? &qso->tx : &qso->rx;

	str_putref(side->county);
	side->county = county;
}

void qso_clear_county(struct qso *qso, bool local)
{
	struct qso_side *side = local ? &qso->tx : &qso->rx;

	str_putref(side->county);
	side->county = NULL;
}

void qso_set_comment(struct qso *qso, struct str *comment)
{
	str_putref(qso->comment);
	qso->comment = comment;
}

void qso_clear_comment(struct qso *qso)
{
	str_putref(qso->comment);
	qso->comment = NULL;
}

void qso_set_zone_continent(struct qso *qso, bool local, const char *zone)
{
	static struct str continents[] = {
		STR_STATIC_INITIALIZER("AF"),
		STR_STATIC_INITIALIZER("AN"),
		STR_STATIC_INITIALIZER("AS"),
		STR_STATIC_INITIALIZER("EU"),
		STR_STATIC_INITIALIZER("NA"),
		STR_STATIC_INITIALIZER("OC"),
		STR_STATIC_INITIALIZER("SA"),
	};
	struct qso_side *side = local ? &qso->tx : &qso->rx;
	size_t i;

	if (!zone)
		goto out;

	for (i = 0; i < ARRAY_LEN(continents); i++) {
		if (strcasecmp(zone, str_cstr(&continents[i])))
			continue;

		/* a real continent */
		str_putref(side->continent);
		side->continent = &continents[i];
		return;
	}

out:
	/* a non-existant continent - clear it */
	str_putref(side->continent);
	side->continent = NULL;
}

void qso_clear_zone_continent(struct qso *qso, bool local)
{
	struct qso_side *side = local ? &qso->tx : &qso->rx;

	str_putref(side->continent);
	side->continent = NULL;
}

int qso_set_zone_dxcc(struct qso *qso, bool local, uint64_t zone)
{
	struct qso_side *side = local ? &qso->tx : &qso->rx;

	/*
	 * Note: in ADIF, code==0 indicates that station is known not to be
	 * within a DXCC entitity
	 */
	val_putref(side->dxcc);
	side->dxcc = VAL_ALLOC_INT(zone);

	return 0;
}

int qso_set_zone_dxcc_user(struct qso *qso, bool local, char *zone)
{
	uint64_t tmp;
	int ret;

	ret = str2u64(zone, &tmp);
	if (!ret)
		ret = qso_set_zone_dxcc(qso, local, tmp);

	return ret;
}

void qso_clear_zone_dxcc(struct qso *qso, bool local)
{
	struct qso_side *side = local ? &qso->tx : &qso->rx;

	val_putref(side->dxcc);
	side->dxcc = NULL;
}

int qso_set_zone_itu(struct qso *qso, bool local, uint64_t zone)
{
	struct qso_side *side = local ? &qso->tx : &qso->rx;

	/* valid zones are [1,90], 0 clears it */
	if (!zone) {
		qso_clear_zone_itu(qso, local);
		return 0;
	} else if ((zone >= 1) && (zone <= 90)) {
		val_putref(side->itu);
		side->itu = VAL_ALLOC_INT(zone);
		return 0;
	} else {
		return -ERANGE;
	}
}

int qso_set_zone_itu_user(struct qso *qso, bool local, char *zone)
{
	uint64_t tmp;
	int ret;

	ret = str2u64(zone, &tmp);
	if (!ret)
		ret = qso_set_zone_itu(qso, local, tmp);

	return ret;
}

void qso_clear_zone_itu(struct qso *qso, bool local)
{
	struct qso_side *side = local ? &qso->tx : &qso->rx;

	val_putref(side->itu);
	side->itu = NULL;
}

int qso_set_zone_cqz(struct qso *qso, bool local, uint64_t zone)
{
	struct qso_side *side = local ? &qso->tx : &qso->rx;

	/* valid zones are [1,40], 0 clears it */
	if (!zone) {
		qso_clear_zone_cqz(qso, local);
		return 0;
	} else if ((zone >= 1) && (zone <= 40)) {
		val_putref(side->cqz);
		side->cqz = VAL_ALLOC_INT(zone);
		return 0;
	} else {
		return -ERANGE;
	}
}

int qso_set_zone_cqz_user(struct qso *qso, bool local, char *zone)
{
	uint64_t tmp;
	int ret;

	ret = str2u64(zone, &tmp);
	if (!ret)
		ret = qso_set_zone_cqz(qso, local, tmp);

	return ret;
}

void qso_clear_zone_cqz(struct qso *qso, bool local)
{
	struct qso_side *side = local ? &qso->tx : &qso->rx;

	val_putref(side->cqz);
	side->cqz = NULL;
}

void qso_set_seq(struct qso *qso, bool local, uint64_t seq)
{
	struct qso_side *side = local ? &qso->tx : &qso->rx;

	val_putref(side->seq);

	/* Note: 0 is a valid sequence number */
	side->seq = VAL_ALLOC_INT(seq);
}

void qso_set_seq_user(struct qso *qso, bool local, char *seq)
{
	uint64_t tmp;

	if (!str2u64(seq, &tmp))
		qso_set_seq(qso, local, tmp);
	else
		qso_set_seq(qso, local, 0);
}

void qso_clear_seq(struct qso *qso, bool local)
{
	struct qso_side *side = local ? &qso->tx : &qso->rx;

	val_putref(side->seq);
	side->seq = NULL;
}

int qso_set(struct qso *qso, enum qso_item what, char *str)
{
	switch (what) {
		case QSO_NOT_A_THING:
			return 0; /* no-op */
		case QSO_UUID:
			if (!xuuid_parse(&qso->uuid, str))
				return -EINVAL;
			xuuid_unparse(&qso->uuid, qso->uuid_str);
			return 0;
		case QSO_START:
		case QSO_END:
			qso_set_datetime(qso, what == QSO_START, str);
			return 0;
		case QSO_COMMENT:
			qso_set_comment(qso, STR_DUP(str));
			return 0;
		case QSO_TX_STATION_CALL:
		case QSO_RX_STATION_CALL:
			qso_set_call(qso, what == QSO_TX_STATION_CALL, str);
			return 0;
		case QSO_TX_OPERATOR_CALL:
		case QSO_RX_OPERATOR_CALL:
			qso_set_operator(qso, what == QSO_TX_OPERATOR_CALL, str);
			return 0;
		case QSO_TX_POWER:
			qso_set_power_user(qso, true, str);
			return 0;
		case QSO_TX_FREQ:
			qso_set_freq(qso, true, str);
			return 0;
		case QSO_TX_MODE:
			qso_set_mode(qso, true, str);
			return 0;
		case QSO_TX_RST:
		case QSO_RX_RST:
			qso_set_rst(qso, what == QSO_TX_RST, str);
			return 0;
		case QSO_RX_NAME:
			qso_set_name(qso, false, STR_DUP(str));
			return 0;
		case QSO_RX_QTH:
			qso_set_qth(qso, false, STR_DUP(str));
			return 0;
		case QSO_TX_GRID:
		case QSO_RX_GRID:
			qso_set_grid(qso, what == QSO_TX_GRID, str);
			return 0;
		case QSO_RX_COUNTRY:
			qso_set_country(qso, false, STR_DUP(str));
			return 0;
		case QSO_RX_STATE:
			qso_set_state(qso, false, STR_DUP(str));
			return 0;
		case QSO_RX_COUNTY:
			qso_set_county(qso, false, STR_DUP(str));
			return 0;
		case QSO_RX_CONTINENT:
			qso_set_zone_continent(qso, false, str);
			return 0;
		case QSO_RX_DXCC:
			qso_set_zone_dxcc_user(qso, false, str);
			return 0;
		case QSO_RX_ITU:
			qso_set_zone_itu_user(qso, false, str);
			return 0;
		case QSO_RX_CQZ:
			qso_set_zone_cqz_user(qso, false, str);
			return 0;
		case QSO_TX_SEQ:
		case QSO_RX_SEQ:
			qso_set_seq_user(qso, what == QSO_TX_SEQ, str);
			return 0;
		case QSO_RX_POWER:
		case QSO_RX_FREQ:
		case QSO_RX_MODE:
		case QSO_TX_NAME:
		case QSO_TX_QTH:
		case QSO_TX_COUNTRY:
		case QSO_TX_STATE:
		case QSO_TX_COUNTY:
		case QSO_TX_CONTINENT:
		case QSO_TX_DXCC:
		case QSO_TX_ITU:
		case QSO_TX_CQZ:
			panic("don't have a good way to set %d", what);
	}

	panic("invalid QSO set item %d", what);
}
