/*
 * Copyright (c) 2020 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#ifndef __QSO_H
#define __QSO_H

#include <jeffpc/error.h>
#include <jeffpc/int.h>
#include <jeffpc/val.h>
#include <jeffpc/uuid.h>
#include <jeffpc/nvl.h>

#include "time.h"

enum qso_item {
	QSO_NOT_A_THING, /* no-op / don't set anything */

	QSO_UUID,
	QSO_START,
	QSO_END,
	QSO_COMMENT,

	QSO_TX_STATION_CALL = 10,
	QSO_TX_OPERATOR_CALL,
	QSO_TX_POWER,
	QSO_TX_FREQ,
	QSO_TX_MODE,
	QSO_TX_RST,
	QSO_TX_NAME,
	QSO_TX_QTH,
	QSO_TX_GRID,
	QSO_TX_COUNTRY,
	QSO_TX_STATE,
	QSO_TX_COUNTY,
	QSO_TX_CONTINENT,
	QSO_TX_DXCC,
	QSO_TX_ITU,
	QSO_TX_CQZ,
	QSO_TX_SEQ,

	QSO_RX_STATION_CALL = 50,
	QSO_RX_OPERATOR_CALL,
	QSO_RX_POWER,
	QSO_RX_FREQ,
	QSO_RX_MODE,
	QSO_RX_RST,
	QSO_RX_NAME,
	QSO_RX_QTH,
	QSO_RX_GRID,
	QSO_RX_COUNTRY,
	QSO_RX_STATE,
	QSO_RX_COUNTY,
	QSO_RX_CONTINENT,
	QSO_RX_DXCC,
	QSO_RX_ITU,
	QSO_RX_CQZ,
	QSO_RX_SEQ,
};

struct qso_side {
	struct str *station_call;
	struct str *operator_call;
	struct val *power; /* mW */
	struct val *freq; /* Hz */
	struct str *mode;
	struct str *rst;
	struct str *name;
	struct str *qth;
	struct str *grid;
	struct str *country;
	struct str *state;
	struct str *county;
	struct str *continent;
	struct val *dxcc;
	struct val *itu;
	struct val *cqz;
	struct val *seq; /* contest qso seq # */
};

struct qso {
	struct xuuid uuid;
	char uuid_str[XUUID_PRINTABLE_STRING_LENGTH];

	struct time start;
	struct time end;
	struct str *comment;

	struct qso_side tx; /* local / tx */
	struct qso_side rx; /* remote / rx */

	/*
	 * additional key-value pairs from ADIF
	 *
	 * If an ADIF field doesn't fit anywhere else in this struct, we
	 * preserve it by storing the unknown fields' values in this nvlist
	 * and writing them out when rewriting the file.
	 */
	struct nvlist *additional;
};

extern int qso_save(const struct qso *qso, int data_dir);
extern void qso_reset(struct qso *qso, bool full);
extern void qso_copy(struct qso *dst, struct qso *src);

/*
 * Set the various parts of a QSO.  All passed in struct str * are consumed.
 */

/* per QSO */
extern int qso_set_datetime(struct qso *qso, bool start, char *str);
extern void qso_set_comment(struct qso *qso, struct str *comment);

/* per side */
extern void qso_set_call(struct qso *qso, bool local, char *call);
extern int qso_set_freq(struct qso *qso, bool local, char *str);
extern void qso_set_grid(struct qso *qso, bool local, char *grid);
extern void qso_set_power(struct qso *qso, bool local, uint64_t milliwatts);
extern int qso_set_power_user(struct qso *qso, bool local, char *power);
extern void qso_set_mode(struct qso *qso, bool local, char *mode);
extern void qso_set_rst(struct qso *qso, bool local, char *rst);
extern void qso_set_name(struct qso *qso, bool local, struct str *name);
extern void qso_set_qth(struct qso *qso, bool local, struct str *qth);
extern void qso_set_country(struct qso *qso, bool local, struct str *country);
extern void qso_set_state(struct qso *qso, bool local, struct str *state);
extern void qso_set_county(struct qso *qso, bool local, struct str *county);
extern void qso_set_zone_continent(struct qso *qso, bool local,
				   const char *zone);
/* valid DXCC entities are [0,inf) */
extern int qso_set_zone_dxcc(struct qso *qso, bool local, uint64_t zone);
extern int qso_set_zone_dxcc_user(struct qso *qso, bool local, char *zone);
/* valid ITU zones are [1,90]; 0 clears the field */
extern int qso_set_zone_itu(struct qso *qso, bool local, uint64_t zone);
extern int qso_set_zone_itu_user(struct qso *qso, bool local, char *str);
/* valid CQ zones are [1,40]; 0 clears the field */
extern int qso_set_zone_cqz(struct qso *qso, bool local, uint64_t zone);
extern int qso_set_zone_cqz_user(struct qso *qso, bool local, char *str);
extern void qso_set_seq(struct qso *qso, bool local, uint64_t seq);
extern void qso_set_seq_user(struct qso *qso, bool local, char *seq);

/*
 * Clear the various parts of a QSO.
 */

/* per QSO */
extern void qso_clear_datetime(struct qso *qso, bool start);
extern void qso_clear_comment(struct qso *qso);

/* per side */
extern void qso_clear_call(struct qso *qso, bool local);
extern void qso_clear_freq(struct qso *qso, bool local);
extern void qso_clear_grid(struct qso *qso, bool local);
extern void qso_clear_power(struct qso *qso, bool local);
extern void qso_clear_mode(struct qso *qso, bool local);
extern void qso_clear_rst(struct qso *qso, bool local);
extern void qso_clear_name(struct qso *qso, bool local);
extern void qso_clear_qth(struct qso *qso, bool local);
extern void qso_clear_country(struct qso *qso, bool local);
extern void qso_clear_state(struct qso *qso, bool local);
extern void qso_clear_county(struct qso *qso, bool local);
extern void qso_clear_zone_continent(struct qso *qso, bool local);
extern void qso_clear_zone_dxcc(struct qso *qso, bool local);
extern void qso_clear_zone_itu(struct qso *qso, bool local);
extern void qso_clear_zone_cqz(struct qso *qso, bool local);
extern void qso_clear_seq(struct qso *qso, bool local);

/* inline helpers */
static inline bool qso_local_item(enum qso_item item)
{
	switch (item) {
		case QSO_NOT_A_THING:
		case QSO_UUID:
		case QSO_START:
		case QSO_END:
		case QSO_COMMENT:
			panic("global qso items aren't local/non-local");

		case QSO_TX_STATION_CALL:
		case QSO_TX_OPERATOR_CALL:
		case QSO_TX_POWER:
		case QSO_TX_FREQ:
		case QSO_TX_MODE:
		case QSO_TX_RST:
		case QSO_TX_NAME:
		case QSO_TX_QTH:
		case QSO_TX_GRID:
		case QSO_TX_COUNTRY:
		case QSO_TX_STATE:
		case QSO_TX_COUNTY:
		case QSO_TX_CONTINENT:
		case QSO_TX_DXCC:
		case QSO_TX_ITU:
		case QSO_TX_CQZ:
		case QSO_TX_SEQ:
			return true;

		case QSO_RX_STATION_CALL:
		case QSO_RX_OPERATOR_CALL:
		case QSO_RX_POWER:
		case QSO_RX_FREQ:
		case QSO_RX_MODE:
		case QSO_RX_RST:
		case QSO_RX_NAME:
		case QSO_RX_QTH:
		case QSO_RX_GRID:
		case QSO_RX_COUNTRY:
		case QSO_RX_STATE:
		case QSO_RX_COUNTY:
		case QSO_RX_CONTINENT:
		case QSO_RX_DXCC:
		case QSO_RX_ITU:
		case QSO_RX_CQZ:
		case QSO_RX_SEQ:
			return false;
	}
}

#endif
