/*
 * Copyright (c) 2020 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <jeffpc/buffer.h>
#include <jeffpc/error.h>
#include <jeffpc/nvl.h>
#include <jeffpc/sexpr.h>

#include "qso_impl.h"

static int append_val(struct buffer *buf, struct val *val,
		      const char *pad, size_t indent);
static int do_unpack(struct nvlist *nvl, struct val *expr);

static int append_atom(struct buffer *buf, struct val *val)
{
	struct str *tmp;
	int ret;

	tmp = sexpr_dump(val, false);

	ret = buffer_append_str(buf, tmp);

	str_putref(tmp);

	return ret;
}

static int append_nvlist(struct buffer *buf, struct nvlist *list, size_t indent)
{
	const struct nvpair *pair;
	bool first;
	char *pad;

	pad = alloca(indent + 1);
	memset(pad, ' ', indent);
	pad[indent] = '\0';

	first = true;
	nvl_for_each(pair, list) {
		struct val *val;

		val = nvpair_value(pair);

#define AW(expr) \
	do { \
		int ret; \
		ret = (expr); \
		if (ret) { \
			val_putref(val); \
			return ret; \
		} \
	} while (0)

		if (first)
			AW(buffer_append_c(buf, '\n'));

		AW(buffer_append_cstr(buf, pad));
		AW(buffer_append_c(buf, '('));
		AW(buffer_append_cstr(buf, nvpair_name(pair)));
		AW(append_val(buf, val, pad, indent));
		AW(buffer_append_cstr(buf, ")\n"));

#undef AW

		val_putref(val);

		first = false;
	}

	return first ? 0 : buffer_append_cstr(buf, pad);
}

static int append_val(struct buffer *buf, struct val *val,
		      const char *pad, size_t indent)
{
	int ret;

	switch (val->type) {
		case VT_STR:
		case VT_INT:
		case VT_NULL:
		case VT_BOOL:
			ret = buffer_append_cstr(buf, " . ");
			if (ret)
				return ret;
			return append_atom(buf, val);
		case VT_NVL:
			return append_nvlist(buf, val_cast_to_nvl(val),
					     indent + 1);
		case VT_ARRAY:
		case VT_SYM:
		case VT_CONS:
		case VT_CHAR:
		case VT_BLOB:
			panic("unsupported val type %d (%s)",
			      val->type, val_typename(val->type));
	}

	return 0;
}

static struct buffer *do_pack(struct nvlist *entry)
{
	struct buffer *buf;
	int ret;

	buf = buffer_alloc(1024);
	if (IS_ERR(buf))
		return buf;

	ret = buffer_append_cstr(buf, "(");
	if (ret)
		goto err;
	append_nvlist(buf, entry, 1);
	if (ret)
		goto err;
	buffer_append_cstr(buf, ")\n");
	if (ret)
		goto err;

	return buf;

err:
	buffer_free(buf);
	return ERR_PTR(ret);
}

struct buffer *qso_pack_sexpr(const struct qso *qso)
{
	struct nvlist *entry;
	struct buffer *buf;

	entry = qso_pack(qso);
	if (IS_ERR(entry))
		return ERR_CAST(entry);

	buf = do_pack(entry);

	nvl_putref(entry);

	return buf;
}

/* set a nested alist */
static int __unpack_cons(struct nvlist *nvl, const char *name, struct val *val)
{
	struct nvlist *child;
	int ret;

	child = nvl_alloc();
	if (IS_ERR(child))
		return PTR_ERR(child);

	ret = do_unpack(child, val);
	if (ret) {
		nvl_putref(child);
		return ret;
	}

	return nvl_set_nvl(nvl, name, child);
}

static int do_unpack(struct nvlist *nvl, struct val *expr)
{
	struct val *cur, *tmp;

	/* top level is made up of a list of cons cells */
	sexpr_for_each_noref(cur, tmp, expr) {
		struct val *name, *val;
		int ret;

		if (cur->type != VT_CONS)
			return -EINVAL;

		name = cur->cons.head;
		val = cur->cons.tail;

		/* name must be a symbol */
		if (name->type != VT_SYM)
			return -EINVAL;

		switch (val->type) {
			case VT_STR:
			case VT_INT:
			case VT_NULL:
				/* (name . val) */
				ret = nvl_set(nvl, val_cstr(name),
					      val_getref(val));
				break;
			case VT_CONS:
				/* (name . (alist)) */
				ret = __unpack_cons(nvl, val_cstr(name), val);
				break;
			case VT_BOOL:
			case VT_CHAR:
			case VT_SYM:
				ret = -EINVAL;
				break;
			case VT_BLOB:
			case VT_ARRAY:
			case VT_NVL:
				panic("unsupported val type %d (%s)",
				      val->type, val_typename(val->type));
		}

		if (ret)
			return ret;
	}

	return 0;
}

int qso_unpack_sexpr(struct qso *qso, const void *buf, size_t len)
{
	struct nvlist *nvl;
	struct val *expr;
	int ret;

	expr = sexpr_parse(buf, len);
	if (IS_ERR(expr))
		return PTR_ERR(expr);

	nvl = nvl_alloc();
	if (IS_ERR(nvl)) {
		ret = PTR_ERR(nvl);
		goto err;
	}

	ret = do_unpack(nvl, expr);
	if (ret)
		goto err_nvl;

	ret = qso_unpack(qso, nvl);

err_nvl:
	nvl_putref(nvl);

err:
	val_putref(expr);

	return ret;
}
