/*
 * Copyright (c) 2020 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <jeffpc/cstr.h>

#include "time.h"

#define TIME_CMP_HELPER(a, b, member) \
	do { \
		if ((a)->member < (b)->member) \
			return -1; \
		if ((a)->member > (b)->member) \
			return +1; \
	} while (0)

int time_cmp(const struct time *a, const struct time *b)
{
	if ((a->have & TIME_PART_DATE) &&
	    (b->have & TIME_PART_DATE)) {
		TIME_CMP_HELPER(a, b, year);
		TIME_CMP_HELPER(a, b, month);
		TIME_CMP_HELPER(a, b, day);
	} else if (a->have & TIME_PART_DATE) {
		return +1;
	} else if (b->have & TIME_PART_DATE) {
		return -1;
	} else {
		/* both mising date - equal so far */
	}

	if ((a->have & TIME_PART_TIME_HM) &&
	    (b->have & TIME_PART_TIME_HM)) {
		TIME_CMP_HELPER(a, b, hour);
		TIME_CMP_HELPER(a, b, min);
	} else if (a->have & TIME_PART_TIME_HM) {
		return +1;
	} else if (b->have & TIME_PART_TIME_HM) {
		return -1;
	} else {
		/* both mising time - equal so far */
	}

	if ((a->have & TIME_PART_TIME_S) &&
	    (b->have & TIME_PART_TIME_S)) {
		TIME_CMP_HELPER(a, b, sec);
	} else if (a->have & TIME_PART_TIME_S) {
		return +1;
	} else if (b->have & TIME_PART_TIME_S) {
		return -1;
	}

	return 0; /* both missing time - actually equal */
}

bool time_snprint_date(char *buf, size_t bufsize, const struct time *t)
{
	if (!t || !(t->have & TIME_PART_DATE)) {
		strcpy_safe(buf, "????""-??""-??", bufsize);

		return false;
	} else {
		snprintf(buf, bufsize, "%04u-%02u-%02u", t->year, t->month,
			 t->day);

		return true;
	}
}

bool time_snprint_time(char *buf, size_t bufsize, const struct time *t)
{
	if (!t || !(t->have & TIME_PART_TIME_HM)) {
		strcpy_safe(buf, "??:??:??", bufsize);

		return false;
	} else {
		if (t->have & TIME_PART_TIME_S)
			snprintf(buf, bufsize, "%02u:%02u:%02u",
				 t->hour, t->min, t->sec);
		else
			snprintf(buf, bufsize, "%02u:%02u:__",
				 t->hour, t->min);

		return true;
	}
}
