/*
 * Copyright (c) 2020 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <math.h>
#include <string.h>

#include <jeffpc/types.h>

#include "util.h"

static int extract_float(const char *input, int64_t *whole_r,
			 int64_t *frac_r, int *scale_r)
{
	uint32_t whole, frac;
	unsigned int scale;
	const char *dp;
	bool negative;
	int ret;

	/* negative sign */
	negative = (*input == '-');
	if ((*input == '+') || (*input == '-'))
		input++;

	/* find the decimal point */
	dp = strchr(input, '.');

	/* whole part */
	ret = str2u32_full(input, &whole, 10, dp ? '.' : '\0');
	if (ret)
		return ret;

	/* fractional part */
	if (dp) {
		ret = str2u32(dp + 1, &frac);
		if (ret)
			return ret;

		scale = pow(10, strlen(dp + 1));
	} else {
		frac = 0;
		scale = 1;
	}

	/* output */
	*whole_r = ((int64_t) whole) * (negative ? -1 : 1);
	*frac_r = ((int64_t) frac) * (negative ? -1 : 1);
	*scale_r = scale;

	return 0;
}

/*
 * Parse a float but return it scaled by the supplied scalar.
 *
 * For example:
 *   - "144.545" with a scale of 1000000 will return 144545000
 *   - "-1.5" with a scale of 3600 will return -5400
 */
int extract_float_as_int64(const char *input, unsigned int scale,
			   int64_t *result_r)
{
	int64_t whole, frac;
	int fscale;
	int ret;

	ret = extract_float(input, &whole, &frac, &fscale);
	if (ret)
		return ret;

	/* TODO: check for overflow */

	/* scale */
	*result_r = (whole * scale) + (frac * scale / fscale);

	return 0;
}

int extract_float_as_int32(const char *input, unsigned int scale,
			   int32_t *result_r)
{
	int64_t tmp;
	int ret;

	ret = extract_float_as_int64(input, scale, &tmp);
	if (ret)
		return ret;

	if ((tmp > INT32_MAX) || (tmp < INT32_MIN))
		return -ERANGE;

	*result_r = tmp;

	return 0;
}

int extract_float_as_double(const char *input, double *result_r)
{
	int64_t whole, frac;
	int scale;
	int ret;

	ret = extract_float(input, &whole, &frac, &scale);
	if (ret)
		return ret;

	/* output */
	*result_r = ((double) whole) + ((double) frac) / scale;

	return 0;
}

bool valid_date(uint32_t y, uint32_t m, uint32_t d)
{
	static const uint8_t month_days[12] = {
		31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31,
	};

	if ((y < 1800) || !m || !d)
		return false;

	if (d > month_days[m - 1])
		return false;

	/* leap year - all good */
	if (((y % 4) == 0) &&
	    (((y % 400) == 0) || ((y % 100) != 0)))
		return true;

	/* non-leap year - Feb 29 is bad */
	return (m != 2) || (d < 29);
}

char *trim_ws(char *str)
{
	char *end;

	if (!str)
		return NULL;

	/* trim leading whitespace */
	str = find_nonws(str);

	/* find last non-whitespace char to keep */
	end = rfind_nonws(str);

	/* null terminate right after - trimming the rest */
	if (end[0] != '\0')
		end[1] = '\0';

	/* return new string */
	return str;
}
